package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;
import edu.fiu.jtlex.timeml.graph.TimeMLGraph;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.*;

/**
 * A Reader Class for initializing Graphs.
 * Takes in either a TimeMLFile, String, or InputStream and creates a corresponding graph.
 *
 * @since jTLEX 1.0
 */
public class GraphReader {
	
	/**
	 * Creates a new empty Graph.
	 * 
	 * @return the created Graph
     * @since jTLEX 1.0
	 */
    public static ITimeMLGraph createTimeMLGraph() {
        TimeMLGraph graph = new TimeMLGraph();
		return graph;
    }
    
    
	/**
	 * Creates a new Graph based on nodes and links.
	 * 
	 * @param nodes	The set of {@link ITimeMLNode}.
	 * @param links	The set of {@link ITimeMLLink}.
	 * @return the created Graph
     * @since jTLEX 1.0
	 */
    public static ITimeMLGraph createTimeMLGraph(Set<ITimeMLNode> nodes, Set<ITimeMLLink> links) {
        TimeMLGraph graph = new TimeMLGraph();
        for (ITimeMLNode node : nodes)
        	graph.addNode(node);
        for (ITimeMLLink link : links)
        	graph.addLink(link);
        return graph;
    }
        
    
	/**
	 * Creates a new Graph from a TimeML file.
	 * 
	 * @param timeMLFile The TimeML file
	 * @return the created Graph
	 * @throws FileNotFoundException If the file does not exist.
     * @throws NullPointerException If the file is null.
     * @since jTLEX 1.0
	 */
    public static ITimeMLGraph createTimeMLGraph(File timeMLFile) throws FileNotFoundException {
        Objects.requireNonNull(timeMLFile);
        InputStream stream = new FileInputStream(timeMLFile);
        TimeMLGraph graph = new TimeMLGraph(stream);
        try {
            stream.close();
        } catch (IOException ignored) {}
		return graph;
    }

	/**
	 * Creates a new Graph from a String.
	 * Cannot be empty or whitespace.
	 * 
	 * @param timeMLString The string of TimeML data.
	 * @return the created Graph
     * @throws NullPointerException If the string is null
     * @since jTLEX 1.0
	 */
    public static ITimeMLGraph createTimeMLGraph(String timeMLString) {
        Objects.requireNonNull(timeMLString);
        InputStream stream = new ByteArrayInputStream(timeMLString.getBytes(StandardCharsets.UTF_8));
        TimeMLGraph graph = new TimeMLGraph(stream);
        try {
            stream.close();
        } catch (IOException ignored) {}
		return graph;
    }
    
	/**
	 * Creates a new Graph from an input stream. The input stream
     * must be closed by the caller.
	 * 
	 * @param timeMLData The InputStream of TimeML data
	 * @return the created Graph
     * @throws NullPointerException If the Stream is null
     * @since jTLEX 1.0
	 */
    public static ITimeMLGraph createTimeMLGraph(InputStream timeMLData) {
        Objects.requireNonNull(timeMLData);
        TimeMLGraph graph = new TimeMLGraph(timeMLData);
        return graph;
    }
    
    

}