package edu.fiu.jtlex.algorithms;

import java.util.List;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.TimeMLTimex;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;

/**
 * Abstraction for a processor to increase connectivity between
 * graph partitions.
 *
 * @since jTLEX 1.0
 */
public interface IDisconnectivityProcessor {

    /**
     * Gets a timex that is the date creation time tag from a partition
     *
     * @param partition The partition from which to extract the timex
     * @return The date creation time timex, or null if there is none
     * @throws NullPointerException If the partition is null
     * @since jTLEX 1.0
     */
    TimeMLTimex getDct(ITimeMLGraph partition);

    /**
     * Determines whether a partition has a date creation time tag
     *
     * @param partition The partition to examine
     * @return True if there is any date creation time tag, else false
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    boolean hasDct(ITimeMLGraph partition);

    /**
     * Determines whether a partition contains any timex
     *
     * @param partition The partition to examine
     * @return True if there is any timex, else false
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    boolean hasTimeExpressions(ITimeMLGraph partition);

    /**
     * Gets a list of time expressions from a partition
     *
     * @param partition The partition from which to extract timexes
     * @return A List of timexes from the partition
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    List<TimeMLTimex> getTimeExpressions(ITimeMLGraph partition);

    /**
     * Suggests a list of possible links to increase connectivity within a graph.
     *
     * @param partitions A List of partitions
     * @param linksCount The number of links in the graph (Used for link creation)
     * @return A List with the suggested links to increase connectivity
     * @throws IllegalArgumentException If the graph does not have a date creation time tag.
     * @since jTLEX 1.0
     */
    List<ITimeMLLink> suggestLinks(List<ITimeMLGraph> partitions, int linksCount);

    /**
     * Gets the quarter the month is bounded to out of the 12 months.
     *
     * @param month Month ranging from <code>1-12</code>
     * @return The quarter the month is bounded to: Q1-Q2-Q3-Q4
     * @throws IllegalArgumentException If the month is out of range
     * @since jTLEX 1.0
     */
    String convertMonthToQuarter(int month);

    /**
     * Gets the half the month is bounded to out of the 12 months.
     *
     * @param month Month ranging from <code>1-12</code>
     * @return The half the month is bounded to: H1-H2.
     * @throws IllegalArgumentException If the month is out of range
     * @since jTLEX 1.0
     */
    String convertMonthToHalf(int month);
}
