package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.graph.IPartitionResult;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;
import edu.fiu.jtlex.timeml.graph.PartitionResult;
import edu.fiu.jtlex.timeml.graph.TimeMLGraph;
import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.SLink;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Class used to partition an input TimeML temporal graph into main subgraph and
 * subordinating subgraphs. TimeML graphs are split by trimming all instances of
 * ILinks that are of type SLinkType and then reassembling the graphs using
 * edges.
 *
 * @since jTLEX 1.0
 */
public class Partitioner implements IPartitioner {

    /**
     * Creates a new partitioner instance
     *
     * @since jTLEX 1.0
     */
    public Partitioner() {}

    /**
     * Partitions a TimeML temporal graph by trimming all S-Links and then
     * reassembling the graphs using DFS.
     *
     * @param timeMLGraph The TimeML graph to be partitioned.
     * @return An IPartitionResult containing the partitions and links.
     * @throws NullPointerException If the graph is null.
     * @since jTLEX 1.0
     */
    @Override
    public IPartitionResult partitionGraph(ITimeMLGraph timeMLGraph) {
        Objects.requireNonNull(timeMLGraph);

        Set<ITimeMLNode> visitedNodes = new HashSet<>();
        List<ITimeMLGraph> subGraphs = new ArrayList<>();
        Set<ITimeMLLink> sLinks = timeMLGraph.getLinks().stream().filter(this::isSLink).collect(Collectors.toSet());
        Set<ITimeMLLink> singleSLinks = new HashSet<>();

        for (ITimeMLLink sLink : sLinks) {
            if (isSingleSLink(sLink, timeMLGraph)) {
                singleSLinks.add(sLink);
            }
            timeMLGraph.removeLink(sLink);
        }

        // Re-assemble the main subgraph/subordinating graphs using DFS
        for (ITimeMLNode node : timeMLGraph.getNodes()) {
            if (!visitedNodes.contains(node)) {
                ITimeMLGraph subgraph = DFS(node, visitedNodes, new TimeMLGraph(), timeMLGraph);
                subGraphs.add(subgraph);
            }
        }

        return new PartitionResult(subGraphs, singleSLinks, timeMLGraph);
    }

    /**
     * Checks whether the given link is S-Link
     *
     * @param link Link to examine
     * @return True if the link is S-Link, else false
     * @throws NullPointerException If the link is null
     * @since jTLEX 1.0
     */
    private boolean isSLink(ITimeMLLink link) {
        Objects.requireNonNull(link);

        return link.getRelType() instanceof SLink;
    }

    /**
     * Determines whether the S-Link is alone for the purpose of partitioning
     *
     * @see #hasCommonInLink(ITimeMLLink)
     * @see #hasCommonOutLink(ITimeMLLink)
     * @throws NullPointerException If the link is null
     * @throws IllegalArgumentException If the link is not an S-Link
     * @since jTLEX 1.0
     */
    private boolean isSingleSLink(ITimeMLLink sLink, ITimeMLGraph mainGraph) {
        Objects.requireNonNull(sLink);

        if (!isSLink(sLink))
            throw new IllegalArgumentException("Expected S-Link, got " + sLink.getRelType());

        return !hasCommonOutLink(sLink, mainGraph) && !hasCommonInLink(sLink, mainGraph);
    }

    /**
     * Determines whether any of the outgoing links from the source node has the
     * same destination node as that of the given link.
     *
     * Example:
     * A --sLink--> B
     * A --Other--> B
     * Returns True
     *
     * @param link Link to examine
     * @return True if such link exists, else false
     * @throws NullPointerException If the link is null
     * @since jTLEX 1.0
     */
    private boolean hasCommonOutLink(ITimeMLLink link, ITimeMLGraph mainGraph) {
        Objects.requireNonNull(link);
        Map<String, Set<ITimeMLLink>> outLinks = mainGraph.getOutLinks();
        Set<ITimeMLLink> startNodeOutLinks = outLinks.get(link.getEventInstance());
        
        return startNodeOutLinks.stream()
                .filter(nodeLink -> !nodeLink.equals(link))
                .anyMatch(nodeLink -> nodeLink.getRelatedToNode().equals(link.getRelatedToNode()));
    }

    /**
     * Determines whether the destination of any incoming link to the source node
     * is the same as the source node of the given link.
     *
     * Example:
     * A --sLink--> B
     * A <--Other-- B
     * Returns True
     *
     * @param link Link to examine
     * @return True if such link exists, else false
     * @throws NullPointerException If the link is null
     * @since jTLEX 1.0
     */
    private boolean hasCommonInLink(ITimeMLLink link, ITimeMLGraph mainGraph) {
        Objects.requireNonNull(link);

        Map<String, Set<ITimeMLLink>> inLinks = mainGraph.getInLinks();
        Set<ITimeMLLink> startNodeInLinks = inLinks.get(link.getEventInstance());

        return startNodeInLinks.stream()
                .filter(nodeLink -> !nodeLink.equals(link))
                .anyMatch(nodeLink -> nodeLink.getEventInstance().equals(link.getRelatedToNode()));
    }

    /**
     * For every node which does not exist in the visitedNodes set:
     * - Add them to the subgraph
     * - Add them to the visitedNodes set
     * - Follow the incoming/outgoing links to the neighboring nodes:
     * - Check if the incoming/outgoing links are already on the current graph, add them if not.
     * - Call the method recursively on the node's neighbors along with the visited  nodes set and the graph.
     *
     * @param node The node to add
     * @param visitedNodes Set of visited nodes
     * @param subgraph The starting point
     * @return The graph's partition
     * @throws NullPointerException If any of the arguments is null
     * @since jTLEX 1.0
     */
    private static ITimeMLGraph DFS(ITimeMLNode node, Set<ITimeMLNode> visitedNodes, ITimeMLGraph subgraph, ITimeMLGraph mainGraph) {
        Objects.requireNonNull(node);
        Objects.requireNonNull(visitedNodes);
        Objects.requireNonNull(subgraph);
        Map<String, Set<ITimeMLLink>> inLinks = mainGraph.getInLinks();
        Map<String, Set<ITimeMLLink>> outLinks = mainGraph.getOutLinks();
        Set<ITimeMLLink> nodeOutLinks = outLinks.get(node.getIdStr());
        Set<ITimeMLLink> nodeInLinks = inLinks.get(node.getIdStr());
        visitedNodes.add(node);
        subgraph.addNode(node);

        for (ITimeMLLink link : nodeOutLinks) {
            if (!subgraph.getLinks().contains(link)) {
            	subgraph.addNode(mainGraph.getNodeById(link.getEventInstance()));
            	subgraph.addNode(mainGraph.getNodeById(link.getRelatedToNode()));
                subgraph.addLink(link);
            }
            if (!visitedNodes.contains(mainGraph.getNodeById(link.getRelatedToNode())))
                DFS(mainGraph.getNodeById(link.getRelatedToNode()), visitedNodes, subgraph, mainGraph);
        }
        for (ITimeMLLink link : nodeInLinks) {
            if (!subgraph.getLinks().contains(link)) {
            	subgraph.addNode(mainGraph.getNodeById(link.getEventInstance()));
            	subgraph.addNode(mainGraph.getNodeById(link.getRelatedToNode()));
                subgraph.addLink(link);
            }
            if (!visitedNodes.contains(mainGraph.getNodeById(link.getEventInstance())))
                DFS(mainGraph.getNodeById(link.getEventInstance()), visitedNodes, subgraph, mainGraph);
        }

        return subgraph;
    }
}
