package edu.fiu.jtlex.algorithms;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Objects;

/**
 * A FileWriter proxy class that implements {@link IWriter}
 *
 * @see IWriter
 * @see FileWriter
 * @since jTLEX 1.0
 */
public class TimeMLWriter implements IWriter {

    private final FileWriter _writer;

    /**
     * Creates a FileWriter proxy.
     *
     * @param filename The name of the file to be created.
     * @throws NullPointerException If filename is null
     * @throws IllegalArgumentException If the given file is a directory
     * @throws IOException If an IO error occurs
     * @since jTLEX 1.0
     */
    public TimeMLWriter(String filename) throws IOException {
        Objects.requireNonNull(filename);

        File file = new File(filename);
        if (file.isDirectory())
            throw new IllegalArgumentException(filename + " is a directory.");

        _writer = new FileWriter(file);
    }

    /**
     * Writes a string
     *
     * @param input String to be written
     * @throws IOException If an IO error occurs
     * @since jTLEX 1.0
     */
    @Override
    public void write(String input) throws IOException {
        _writer.write(input);
    }

    /**
     * Writes a string and a new line character at the end.
     *
     * @param input String to be written
     * @throws IOException If an IO error occurs
     * @since jTLEX 1.0
     */
    @Override
    public void writeLine(String input) throws IOException {
        _writer.write(input);
        _writer.write("\n");
    }

    /**
     * Closes the stream.
     *
     * @throws IOException If an IO error occurs
     * @since jTLEX 1.0
     */
    @Override
    public void close() throws IOException {
        _writer.close();
    }
}
