package edu.fiu.jtlex.timeml.data;

import edu.fiu.jtlex.timeml.graph.ITimeMLNode;

/**
 * Interface of Link attributes shared by 3 Links (TLink,SLink,ALink).
 * @since jTLEX 1.0
 * @author Pablo Maldonado
 *
 */


public interface ITimeMLLink {
	/**
	 * Interface of to propagate children.
	 * Similar in Implementation to {@link ITimeMLNode.IType}. The different types of Links are 
	 * represented by enums in {@link ITimeMLLink} such as TLinks or ALinks. In order for them
	 * to be grabbed in a link when it could be either one, they are utilized by this wrapper interface
	 * so it returns the same type.
	 * @since jTLEX 1.0
	 * @author Stephan Belizaire
	 * 
	 */
	public interface ITimeMLLinkType {
		
	}
		
	/**
	 * Enum that represents temporal relation holding between
	 * the entities for a TLink.<p>
	 * Possible values are: <ul><li>BEFORE</li><li>AFTER</li><li>INCLUDES</li>
	 * <li>IS INCLUDED</li><li>DURING</li><li>DURING INV</li><li>SIMULTANEOUS</li>
	 * <li>IAFTER</li><li>IBEFORE</li><li>IDENTITY</li><li>BEGINS</li><li>ENDS</li>
	 * <li>BEGUN BY</li><li>ENDED BY</li></ul>
	 */

	public enum TLink implements ITimeMLLinkType{
		/**
		 * One before the other. <br>
		 * As in the following example between the events slayings and arrested:
		 * <ul><li>The police looked into the <b>slayings</b> of 14 women. In six of
		 * the cases suspects have already been <b>arrested</b>.
		 * </ul>
		 */
		BEFORE,
		
		/**
		 * One after the other. <br>
		 * Inverse of {@link TLink#BEFORE}. <br>
		 * So the two events examples in
		 * the {@link TLink#BEFORE} documentation can alternatively be annotated as expressing an
		 * after relation, if the directionality is changed.
		 */
		AFTER,
		
		/**
		 * One including the other. <br>
		 * As is the case between the temporal expression and the event in the
		 * following example:
		 * <ul><li>John <b>arrived</b> in Boston last <b>Thursday</b>.
		 * </ul>
		 */
		INCLUDES,
		
		/**
		 * One being included in the other.
		 * Inverse of {@link TLink#INCLUDES}.
		 */
		IS_INCLUDED,
		
		/**
		 * One holds during the other. <br>
		 * Specifically applicable to states or events that persist throughout a
		 * duration, for example:
		 * <ul>
		 * <li>James was <b>CTO</b> for <b>two years</b>.
		 * <li>John <b>taught</b> for <b>20 minutes</b> on Monday.
		 * </ul>
		 */
		DURING,
		
		/**
		 * One is being held during the other. 
		 * Inverse of {@link TLink#DURING}.
		 */
		DURING_INV,
		
		/**
		 * Two event instances are judged simultaneous if they happen at the
		 * same time, or are temporally indistinguishable in context. I.e. occur
		 * close enough to the same time that further distinguishing their times
		 * makes no difference to the temporal interpretation of the text.
		 */
		SIMULTANEOUS,
		
		/**
		 * One immediately after than the other.
		 * Inverse of {@link TLink#IBEFORE}.
		 */
		IAFTER,
		
		/**
		 * One immediately before the other. <br>
		 * As in the following sentence between <i>crash</i> and <i>died</i>.
		 * <ul><li>All passengers <b>died</b> when the plane <b>crashed</b> 
		 * 		into the mountain.
		 * </ul>
		 */
		IBEFORE,
		
		/** Event identity. <br>
		 * For example:
		 * <ul>
		 * <li> John <b>drove</b> to Boston. During his <b>drive</b> he ate a donut.
		 * </ul>
		 */
		IDENTITY,
		
		/**
		 * One being the beginning of the other. <br>
		 * As holds between the first of the temporal expressions and the event
		 * in the following example:
		 * <ul>
		 * <li>John was <b>in the gym</b> between <b>6:00 p.m.</b> and 7:00 p.m.
		 * </ul>
		 */
		BEGINS,
		
		/**
		 * One being the ending of the other. <br>
		 * For example:
		 * <ul>
		 * <li>John was <b>in the gym</b> between 6:00 p.m. and <b>7:00 p.m.</b>
		 * </ul>
		 */
		ENDS,
		
		/**
		 * One being begun by the other. 
		 * Inverse of {@link TLink#BEGINS}.
		 */
		BEGUN_BY,
		
		/**
		 * One being ended by the other.
		 * Inverse of {@link TLink#ENDS}.
		 */
		ENDED_BY
	}

	/**
	 * Enum that represents the temporal relation holding between
	 * the entities for ALink.<p>
	 * Possible values are: <ul><li>INITIATES</li><li>CULMINATES</li>
	 * <li>TERMINATES</li><li>CONTINUES</li><li>REINITIATES</li></ul>
	 */
	public enum ALink implements ITimeMLLinkType{
		/**
		 * John <b>started</b> to <i>read</i>.
		 */
		INITIATES,
		
		/**
		 * John <b>finished</b> <i>assembling</i> the table.
		 */
		CULMINATES,
		
		/**
		 * John <b>stopped</b> <i>talking</i>.
		 */
		TERMINATES,
		
		/**
		 * John <b>kept</b> <i>talking</i>.
		 */
		CONTINUES,
		
		/**
		 * John <b>resumed</b> <i>talking</i>.
		 */
		REINITIATES
		
	}
	
	
	/**
	 * Enum that represents the temporal relation holding between
	 * the entities for SLink.
	 * Possible values are: <ul><li>MODAL</li><li>EVIDENTIAL</li>
	 * <li>NEG EVIDENTIAL</li><li>FACTIVE</li><li>COUNTER FACTIVE</li>
	 * <li>CONDITIONAL</li></ul>
	 */
	public enum SLink implements ITimeMLLinkType{
		/**
		 * This relation is brought up by events introducing a reference to a
		 * possible world, mainly I_ACTIONs and I_STATEs.
		 * <ul>
		 * <li>John <b>promised</b> Mary to buy some beer.
		 * <li>Mary <b>wanted</b> John to buy some wine.
		 * </ul>
		 */
		MODAL,
		
		/**
		 * Evidential relations are typically introduced by REPORTING or PERCEPTION
		 * events.
		 * <ul>
		 * <li>John <b>said</b> he bought some wine.
		 * <li>Mary <b>saw</b> John carrying only beer.
		 * </ul>
		 */
		EVIDENTIAL,
		
		/**
		 * Introduced by REPORTING and PERCEPTION events conveying
		 * negative polarity.
		 * <ul>
		 * <li>John <b>denied</b> he bought only beer.
		 * </ul>
		 */
		NEG_EVIDENTIAL,
		
		/**
		 * Certain verbs introduce an entailment (or presupposition) of their argument's
		 * veracity. They include forget (with a tensed complement),
		 * regret, or manage:
		 * <ul>
		 * <li>John <b>forgot</b> that he was in Boston last year.
		 * <li>Mary <b>regrets</b> that she didn't marry John.
		 * <li>John <b>managed</b> to leave the party.
		 * </ul>
		 */
		FACTIVE,
		
		/**
		 * The event introduces a 
		 * presupposition about the non-veracity of its argument: forget (to),
		 * unable to (in past tense), prevent, cancel, avoid, decline, etc.
		 * <ul>
		 * <li>John <b>forgot</b> to buy some wine.
		 * <li>Mary was <b>unable</b> to marry John.
		 * <li>John <b>prevented</b> the divorce.
		 * </ul>
		 */
		COUNTER_FACTIVE,
		
		/**
		 * A conditional relationship can occur between any two event instances
		 * and is generally accompanied by a signal such as if. The antecdent of
		 * the conditional takes the place of the introducing event instance (in
		 * bold) and the consequent takes the place of the subordinated event
		 * instance (underlined):
		 * <ul>
		 * <li><i>If</i> John <b>brings</b> the beer, Mary will bring the chips.
		 * </ul>
		 */
		CONDITIONAL
	}
	
	
	
	/**
	 * Enum that represents whether the TLink was generated manually or by closure.
	 * 2 options: <ul><li>{@code MANUALLY}</li><li>{@code CLOSURE}</li></ul>
	 */
	public static enum Origintype{
		/**
		 * TODO: Is this constant necessary?
		 */
		USER,
		
		/**
		 * Link was manually generated.
		 */
		MANUALLY,
		
		/**
		 * Link was generated in the closure part of the annotation.
		 */
		CLOSURE
	}
	
	
	
	/**
	 * Returns the ID of the link. <br>
	 * Link ID gives each TimeML link an identification number
	 * OPTIONAL
	 * @since jTLEX 1.0
	 * @return the ID of the link.
	 */
	public int getLinkID();
	
	/**
	 * Returns the type of Link.
	 * @since jTLEX 1.0
	 * @return {@link ITimeMLLink.TLink} / {@link ITimeMLLink.ALink} / {@link ITimeMLLink.SLink}
	 */
	public Class<ITimeMLLink> getLinkTag();
	
	/**
	 * Returns the syntax of the link. <br>
	 * Syntax holds any syntactic pattern that was used in generating the link (e.g.
	 * from an automatic annotation tool). <br>
	 * OPTIONAL
	 * @since jTLEX 1.0
	 * @return the syntax of the link.
	 */
	public String getSyntax();
	
	/**
	 * Returns the ID of the event instance involved in the link.
	 * OBLIGATORY
	 * @since jTLEX 1.0
	 * @return the ID of the event instance involved in the link.
	 */
	public String getEventInstance();
	
	
	/**
	 * Returns the ID of the relation signal. <br>
	 * If the temporal relation holding between the entities
	 * is explicitly signaled in the text, then the ID of that signal needs
	 * to be supplied here.<br>
	 * OPTIONAL
	 * @since jTLEX 1.0
	 * @return the optional ID of the relation signal.
	 */
	public ITimeMLSignal getSignal();
	
	
	/**
	 * Returns whether a TLINK was generated 
	 * {@code MANUALLY} or by {@code CLOSURE}. <br>
	 * THIS IS ONLY FOR TLINK
	 * @since jTLEX 1.0
	 * @return Whether link was generated {@code MANUALLY} or by {@code CLOSURE}
	 */
	public Origintype getOrigin();
	
	/**
	 *  
	 * Returns the ID of the entity that is being related to the event instance
	 * with ID=eventInstanceID or time expression with ID=timeID. <br>
	 * relatedToEventInstance or relatedToTime needs to be present.
	 * @since jTLEX 1.0
	 * @return the ID of the entity that is being related to the event instance
	 * with ID=eventInstanceID or time expression with ID=timeID.
	 */
	public String getRelatedToNode();
	
	
	/**
	 * This is the temporal relation holding between
	 * the entities. <br>These are represented by 3 different enums:
	 * <ul>
	 * <li>ALinkRelType
	 * <li>SLinkRelType
	 * <li>TLinkRelType
	 * </ul>
	 * <br>The returned type depends on the linkTag.
	 * @see ALink
	 * @see SLink
	 * @see TLink
	 * @since jTLEX 1.0
	 * @return {@code ALinkRelType} / {@code SLinkRelType} / {@code TLinkRelType}
	 */
	public ITimeMLLinkType getRelType();
	
	
	/**
	 * Returns the String representation of the Link.
	 * @since jTLEX 1.0
	 * @return the String representation of the Link.
	 */
	public String toString();
	
	/**
	 * Returns the JSON (RFC 8259) representation of the Link.
	 * @since jTLEX 1.0
	 * @return the JSON (RFC 8259) representation of the Link.
	 */
	public String toJson();
	
	/**
	 * Returns a deep copy of the Link.
	 * @since jTLEX 1.0
	 * @return  a deep copy of the Link.
	 */
	public ITimeMLLink clone();
	
    /**
	 * Generates a hash code for the link.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
     */
    public int hashCode();
	
	/**
	 * Checks if this link equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both links are equal, i.e. same ID and tag, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);
}
