package edu.fiu.jtlex.timeml.data;

/**
 * Represents the signal tag.<p>
 * A signal is a textual element that makes explicit the relation holding between
 * two entities (timex and event, timex and timex, or event and event). Signals
 * are generally:
 * <ul>
 * <li>Temporal prepositions: on, in, at, from, to, before, after, during, etc.
 * <li>Temporal conjunctions: before, after, while, when, etc.
 * <li>Prepositions signaling modality: to.
 * <li>Special characters: X-X and X/X, in temporal expressions denoting
 * 		ranges (September 4-6, Apr. 1999/Jul. 1999, etc.).
 * </ul>
 * @since jTLEX 1.0
 * @author arada002
 *
 */
public interface ITimeMLSignal {
	
	/**
	 * Returns a unique signal ID number.
	 * @return a unique signal ID number. 
	 * 		   May not return less than 1
	 * @since jTLEX 1.0
	 */
	public int getID();
	
	/**
	 * Returns a unique signal ID string.
	 * @return a unique signal ID string. 
	 * 		   May not return  <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public String getIdStr();
	
	/**
	 * Returns the temporal preposition, conjunction or modifier phrase.
	 * @return the temporal preposition, conjunction or modifier phrase. 
	 * 		   May not return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public String getSignalString();
	
	/**
	 * Returns the String representation of the Signal.
	 * @return the String representation of the Signal.
	 * @since jTLEX 1.0
	 */
	public String toString();
	
	/**
	 * Returns the JSON (RFC 8259) representation of the Signal.
	 * @return the JSON (RFC 8259) representation of the Signal.
	 * @since jTLEX 1.0
	 */
	public String toJson();
	
	/**
	 * Returns a deep copy of the Signal.
	 * @return a deep copy of the Signal.
	 * @since jTLEX 1.0
	 */
	public ITimeMLSignal clone();
	
	
	/**
	 * Generates a hash code for the signal.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode();
    
	/**
	 * Checks if this Signal equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both signals are equal, i.e. same ID and string, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);
	
}
