package edu.fiu.jtlex.timeml.data;

import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

/**
 * Default implementation of {@code ITimeMLEvent}.
 * @since jTLEX 1.0
 * @author arada002
 */
public class TimeMLEvent implements ITimeMLEvent {

	private final int eID;
	private final EventClass eventClass;
	private final String stem;
	private Set<ITimeMLInstance> instances;

	/**
	 * Creates a new Event with the specified information.
	 * 
	 * @param eID        id of the event; may not be 0 or negative, must be an integer
	 * @param eventClass the class of the entry may not be <code>null</code>, must be an EventClass enum
	 * @param stem       the stem attribute of the event; may not be empty string or
	 *                   all whitespace; may be <code>null</code> if stem is not
	 *                   included in annotation
	 * @throws NullPointerException     if eventClass argument is <code>null</code>
	 * @throws IllegalArgumentException if the stem is empty or all whitespace or
	 *                                  eID is 0/negative/not an integer or if the eventClass is not an enum.
	 */
	public TimeMLEvent(int eID, EventClass eventClass, String stem) {
		if (eID < 1) {
			throw new IllegalArgumentException("EventID can not be less 1.");
		}
		if (eventClass == null) {
			throw new NullPointerException("Event class needs to be defined.");
		}
		if (stem != null) {
			stem = stem.trim();
			if (stem.length() == 0) {
				throw new IllegalArgumentException("Stem can not be empty string or all whitespace");
			}
		}

		this.eID = eID;
		this.eventClass = eventClass;
		this.stem = stem;
		this.instances = new HashSet<ITimeMLInstance>();
	}
	
	/**
	 * Returns the eID of an Event.
	 * eID must always be an integer greater than 0.
	 * Will never return null as the Event constructor requires a valid eID for
	 * the object to be created.
	 * @since jTLEX 1.0
	 * @return the eID of an Event. 
	 */
	public int getId() {
		return eID;
	}

	/**
	 * Grabs the eID, adds an e to the front for String format and returns.
	 * Will never return null as the Event constructor requires a valid eID for 
	 * the object to be created.
	 * @since jTLEX 1.0
	 * @return the eID of an Event in String format. 
	 */
	public String getIdStr() {
		return "e" + eID;
	}
	
	/**
	 * Returns the stem of an Event.
	 * Will never return null as the Event constructor checks for a valid stem while
	 * creating the Event object.
	 * @since jTLEX 1.0
	 * @return the stem of an Event. 
	 */
	public String getStem() {
		return stem;
	}

	/**
	 * Returns the event class of an Event.
	 * Will never return null as the Event Constructor requires an Event Class while 
	 * creating the object.
	 * @since jTLEX 1.0
	 * @return the event class of an Event. 
	 */
	public EventClass getEventClass() {
		return eventClass;
	}

	/**
	 * Returns the info of an Event in String Format.
	 * Will never return a null string.
	 * @since jTLEX 1.0
	 * @return the info of an Event in String Format. 
	 */
	@Override
	public String toString() {
		return "EVENT: eid = e" + getId() + ", class = " + getEventClass().toString() + ", stem = "+ getStem();
	}
	
	
	/**
	 * Sets the related Instance for the event.
	 * May be null.
	 * 
	 * @since jTLEX 1.0
	 */
	public void setInstance(ITimeMLInstance instance) {
		this.instances.add(instance);
	}
	
	/**
	 * Returns the IDs of the instances related to the Event.
	 * May be null
	 * @since jTLEX 1.0
	 * @return the int ID of the Instances related to the event. 
	 */
	public Set<Integer> getInstances() {
		Set<Integer> instanceIDs = new HashSet<Integer>();
		for (ITimeMLInstance instance : this.instances)
			instanceIDs.add(instance.getId());
		return instanceIDs;
	}
	
	/**
	 * Converts the Event and it's info into JSON Format.
	 * Will never return a null string.
	 * @since jTLEX 1.0
	 * @return the JSON format of an Event. 
	 */
	public String toJson() {
		return "{\"id\":\"" + this.getIdStr() + "\", " +
				"\"eventClass\":\"" + this.getEventClass().toString() + "\", " +
				"\"stem\":\"" + this.getStem() + "\"}";
	}
	
	/**
	 * Creates and returns a field-by-field copy of the Event.
	 * Will never return a null or empty Event object.
	 * @since jTLEX 1.0
	 * @return a copy of an Event. 
	 */
	public TimeMLEvent clone() {
		TimeMLEvent clone = new TimeMLEvent(this.eID, this.eventClass, this.stem);
		for (ITimeMLInstance instance : this.instances) {
			clone.setInstance(instance);
		}
		return clone;
	}
	
	/**
	 * Generates a hash code for the event.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode() {
        int hashCode = 1;
        hashCode = 31*hashCode + this.eID;
        hashCode = 31*hashCode + (this.eventClass==null ? 0 : this.eventClass.hashCode());
        hashCode = 31*hashCode + (this.stem==null ? 0 : this.stem.hashCode());
        return hashCode;
    }
    
	/**
	 * Checks if this event equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both events are equal, i.e. same ID and eventClass, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLEvent))
            return false;

        ITimeMLEvent obj = (TimeMLEvent) o;

        if (this.eID != obj.getId())
            return false;
        
        if (this.getEventClass() != obj.getEventClass())
        	return false;

        if (this.stem != obj.getStem())
        	return false;
        
        if (!this.getInstances().equals(obj.getInstances()))
        	return false;


        return true;
    }

}
