package edu.fiu.jtlex.timeml.data;

import java.util.Objects;

import edu.fiu.jtlex.timeml.data.ITimeMLEvent.EventClass;

/**
 * Default implementation of {@code IMakeInstance}.
 * @since jTLEX 1.0
 * @author arada002
 */

public class TimeMLInstance implements ITimeMLInstance {

    private final int eventInstanceID;
    private final ITimeMLEvent event;
    private final Tense tense;
    private final Aspect aspect;
    private final Pos pos;
    private final Polarity polarity;
    private final String modality;
    private final ITimeMLSignal signal;
    private final String cardinality;

    /**
     * Creates a new event Instance object.
     *
     * @param eventInstanceID a unique ID number
     *                        may not be less than 1
     * @param event           the even to which event instance is related
     * @param tense           phrase tense
     *                        may not be <code>null</code>
     * @param aspect          phrase aspect
     *                        may not be <code>null</code>
     * @param pos             part of speech, syntactic category of phrase
     *                        may not be <code>null</code>
     * @param polarity        indicates whether the eventInstance phrase is negated
     *                        may not be <code>null</code>
     * @param modality        modal word that modifies the instance
     *                        may be <code>null</code>
     * @param signal          the signal object, present if multiple event instances are created
     *                        <code>null</code> if not defined
     *                        may not be less than 0
     * @param cardinality     only present if there is a signal that provides its value
     *                        0 if not defined
     *                        may not be less than 0
     *                        
     * @throws NullPointerException if Event, Tense, Aspect, Pos, or Polarity are null
     * 
     * @throws IllegalArgumentException if Modality is empty or all whitespace
     */
    public TimeMLInstance(int eventInstanceID, ITimeMLEvent event, Tense tense, Aspect aspect, Pos pos,
                    Polarity polarity, String modality, ITimeMLSignal signal, String cardinality) {

        if (eventInstanceID < 1) {
            throw new IllegalArgumentException("Event instance ID can not be less than 1.");
        }


        if (event == null) {
            throw new NullPointerException("Event can not be null");
        }

        if (tense == null) {
            throw new NullPointerException("Tense has to be defined.");
        }

        if (aspect == null) {
            throw new NullPointerException("Aspect has to be defined.");
        }

        if (pos == null) {
            throw new NullPointerException("Part of speech has to be defined.");
        }

        if (polarity == null) {
            throw new NullPointerException("Polarity has to be defined.");
        }

        if (modality != null) {
            modality = modality.trim();
            if (modality.length() == 0) {
                throw new IllegalArgumentException("Modality can not be empty string or "
                        + "all whitespace");
            }
        }

        this.eventInstanceID = eventInstanceID;
        this.event = event;
        this.tense = tense;
        this.aspect = aspect;
        this.pos = pos;
        this.polarity = polarity;
        this.modality = modality;
        this.signal = signal;
        this.cardinality = cardinality;

        event.setInstance(this);
    }

    /**
     * Returns the unique ID number eventInstanceID.
     *
     * @return the unique ID number eventInstanceID.
     * @since jTLEX 1.0
     */
    public int getId() {
        return eventInstanceID;
    }

    /**
     * Returns the ID of the event to which the instance is related.
     *
     * @return the ID of the event to which the instance is related.
     * @since jTLEX 1.0
     */
    public String getEvent() {
        return event.getIdStr();
    }

    /**
     * Returns the tense of the verbal phrase.
     *
     * @return the tense of the verbal phrase.
     * May not return <code>null</code>.
     * @since jTLEX 1.0
     */
    public Tense getTense() {
        return tense;
    }

    /**
     * Returns the aspect of the verbal phrase.
     *
     * @return the aspect of the verbal phrase.
     * May not return <code>null</code>.
     * @since jTLEX 1.0
     */
    public Aspect getAspect() {
        return aspect;
    }

    /**
     * Returns the syntactic category of the phrase (part of speech).
     *
     * @return the syntactic category of the phrase (part of speech).
     * May not return <code>null</code>.
     * @since jTLEX 1.0
     */
    public Pos getPos() {
        return pos;
    }

    /**
     * The modality attribute is only present if there is a modal
     * word present that modifies the instance.
     *
     * @return the modal word that modifies the instance.
     * May return <code>null</code>.
     * @see <a href="https://en.wikipedia.org/wiki/Modality_(natural_language)">
     * Modality (natural language)</a>
     * @since jTLEX 1.0
     */
    public String getModality() {
        return modality;
    }

    /**
     * Returns the polarity of the eventInstance.
     * The polarity of an event instance is a required attribute
     * represented by the boolean attribute, polarity. <br>If it is set to
     * NEG, the event instance is negated. <br>If it is set to
     * POS or not present in the annotation, the event instance is not
     * negated.
     *
     * @return polarity:
     * NEG if the phrase is negated,
     * POS if the phrase is not negated.
     * May not return <code>null</code>.
     * @since jTLEX 1.0
     */
    public Polarity getPolarity() {
        return polarity;
    }

    /**
     * Returns the signal object, present if multiple event instances are created
     * <code>null</code> if not defined, may not be less than 0
     *
     * @return the signal object, present if multiple event instances are created.
     * @since jTLEX 1.0
     */
    public ITimeMLSignal getSignal() {
        return signal;
    }

    /**
     * Returns the cardinality, only present if there is a signal that provides its value.
     * 0 if not defined, may not be less than 0.
     *
     * @return the cardinality, only present if there is a signal that provides its value.
     * @since jTLEX 1.0
     */
    public String getCardinality() {
        return cardinality;
    }

    /**
     * Calls the event to which the instance is related, then returns the event class of the event instance.
     *
     * @return the event class of the event instance.
     * @since jTLEX 1.0
     */
    public EventClass getType() {
        return this.event.getEventClass();
    }

    /**
     * Grabs the eventInstanceID, adds eiid to the front for the String format and returns.
     *
     * @return the eventInstanceID in string format.
     * @since jTLEX 1.0
     */
    public String getIdStr() {
        return "eiid" + eventInstanceID;
    }

    /**
     * Returns the info of the Instance in String format.
     *
     * @return the info of the Instance in String format.
     * @since jTLEX 1.0
     */
    @Override
    public String toString() {
        String ret = "Event Instance: \n\t{" +
                "ID = " + this.getIdStr() +
                ", Tense = " + this.getTense() +
                ", Aspect = " + this.getAspect() +
                ", Part of Speech = " + this.getPos() +
                ", Polarity = " + this.getPolarity() +
                ", Modality = \"" + this.getModality() + "\"" +
                ", Cardonality = \"" + this.getCardinality() + "\"";
        if (this.getSignal() == null) {
            ret += ", Signal = null";
        } else {
            ret += "\n\t\t" + this.getSignal();
        }
        ret += "\n\t\t" + this.event.toString() + "\n\t}";

        return ret;
    }

    /**
     * Converts the Instance and its info into JSON format.
     *
     * @return the JSON format of the Instance.
     * @since jTLEX 1.0
     */
    public String toJson() {
        String ret = "{\"id\":\"" + this.getIdStr() +
                "\", \"tense\":\"" + this.getTense() +
                "\", \"aspect\":\"" + this.getAspect() +
                "\", \"partOfSpeech\":\"" + this.getPos() +
                "\", \"polarity\":\"" + this.getPolarity() +
                "\", \"modality\":\"" + this.getModality() +
                "\", \"cardinality\":\"" + this.getCardinality() +
                "\", \"signal\":";

        if (this.getSignal() != null) ret += this.getSignal().toJson();
        else ret += "{}";

        ret += ", \"event\":" + this.event.toJson() + "}";

        return ret;
    }

    /**
     * Creates and returns a field-by-field copy of the Instance.
     *
     * @return a copy of the Instance.
     * @since jTLEX 1.0
     */
    public TimeMLInstance clone() {
        return new TimeMLInstance(
                this.getId(),
                this.event.clone(),
                (ITimeMLInstance.Tense) this.getTense(),
                (ITimeMLInstance.Aspect) this.getAspect(),
                (ITimeMLInstance.Pos) this.getPos(),
                (ITimeMLInstance.Polarity) this.getPolarity(),
                this.getModality(),
                this.getSignal() == null ? null : this.getSignal().clone(),
                this.getCardinality());
    }
    
    
	/**
	 * Checks if this instance equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both instances are equal, i.e. same ID and event, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLInstance))
            return false;

        ITimeMLInstance obj = (TimeMLInstance) o;

        if (this.eventInstanceID != obj.getId())
            return false;
        
        if (!this.getEvent().equals(obj.getEvent()))
        	return false;

        if (this.tense != obj.getTense())
        	return false;
        
        if (this.aspect != obj.getAspect())
        	return false;

        if (this.pos != obj.getPos())
        	return false;
        
        if (this.polarity != obj.getPolarity())
        	return false;
        
        if (this.modality != obj.getModality())
        	return false;
        
        if (this.signal != obj.getSignal())
        	return false;
        
        if (this.cardinality != obj.getCardinality())
        	return false;
        
        return true;
    }

}

