package edu.fiu.jtlex.timeml.data;

import java.util.Objects;

import edu.fiu.jtlex.timeml.graph.ITimeMLNode;

/**
 * Default implementation of {@code ILink}.
 * @since jTLEX 1.0
 * @author Pablo Maldonado
 */

public class TimeMLLink implements ITimeMLLink {

    private final int linkID;
    private final String syntax;
    private final ITimeMLNode eventInstance;
    private final ITimeMLSignal signal;
    private final ITimeMLNode relatedToNode;
    private final Origintype orig;
    protected final Class<ITimeMLLink> linktag;
    private final ITimeMLLinkType relType;

    /**
     * This is the constructor for Tlink.
     *
     * @param linktag       linkTag is the type of Link: TLINK
     * @param linkID        Link ID gives each TimeML link an identification
     *                      number
     * @param syntax        syntax holds any syntactic pattern that was used
     *                      in generating the link
     * @param eventInstance Obligatory attribute (one or the other of these
     *                      needs to be present). This is the ID of the
     *                      eventInstance or the timeID involved in the
     *                      temporal link.
     * @param signal        Optional attribute. If the temporal relation
     *                      holding between the entities is explicitly
     *                      signaled in the text, then the ID of that signal
     *                      needs to be supplied here.
     * @param relatedToNode Obligatory attribute (one or the other of these
     *                      needs to be present). This is the ID of the
     *                      entity that is being related to the event
     *                      instance with ID=eventInstanceID or time
     *                      expression with ID=timeID.
     * @param relType       Obligatory attribute. This is the temporal
     *                      relation holding between the entities.
     * @param orig          Optional attribute. This attribute shows whether
     *                      a TLINK was generated manually or by closure.
     * @throws NullPointerException     if TLinkRelType argument is null
     * @throws IllegalArgumentException throw error if linktag is not the TLink Class
     */

    @SuppressWarnings("unchecked")
	public TimeMLLink(Class<?> linktag, int linkID, String syntax, ITimeMLNode eventInstance, ITimeMLSignal signal, ITimeMLNode relatedToNode,
                ITimeMLLinkType relType, Origintype orig) {

        if (!linktag.equals(ITimeMLLink.TLink.class)) {
            String errorMessage = "Error: using wrong constructor, this is Tlink constructor)";
            throw new IllegalArgumentException(errorMessage);
        }
        if (relType == null || eventInstance == null || relatedToNode == null) {
            throw new NullPointerException();
        }

        this.linktag = (Class<ITimeMLLink>) linktag;
        this.linkID = linkID;
        this.syntax = syntax;
        this.eventInstance = eventInstance;
        this.signal = signal;
        this.relatedToNode = relatedToNode;
        this.relType = relType;
        this.orig = orig;
    }

    /**
     * This is the constructor for Alink and SLink
     *
     * @param linktag       linkTag is the type of Link: SLINK or ALINK
     * @param linkID        Link ID gives each TimeML link an
     *                      identification number
     * @param syntax        syntax holds any syntactic pattern that was
     *                      used in generating the link
     * @param eventInstance Obligatory attribute. This is the ID of the
     *                      (aspectual) eventInstance involved in the
     *                      aspectual link.
     * @param signal        Optional attribute. If the aspectual relation
     *                      holding between the events is explicitly
     *                      signaled in the text, then the ID of that
     *                      signal needs to be filled in here
     * @param relatedToNode Obligatory attribute. This is the ID of the
     *                      event instance related to the aspectual event.
     * @param relType       Obligatory attribute. This is the temporal
     *                      relation holding between the entities for Alink
     * @throws NullPointerException     if ALinkRelType argument is null
     * @throws IllegalArgumentException throw error if it is not an ALink or SLink class
     */
    @SuppressWarnings("unchecked")
	public TimeMLLink(Class<?> linktag, int linkID, String syntax, ITimeMLNode eventInstance, ITimeMLSignal signal, ITimeMLNode relatedToNode,
                ITimeMLLinkType relType) {
    	if (!(linktag.equals(ITimeMLLink.ALink.class) || (linktag.equals(ITimeMLLink.SLink.class)))) {

      	  String errorMessage = "Error: using wrong constructor, this is Alink or SLink constructor)";
            throw new IllegalArgumentException(errorMessage);
    	} 
        if (relType == null || eventInstance == null || relatedToNode == null) {
            throw new NullPointerException();
        }

        this.linktag = (Class<ITimeMLLink>) linktag;
        this.linkID = linkID;
        this.syntax = syntax;
        this.eventInstance = eventInstance;
        this.signal = signal;
        this.relatedToNode = relatedToNode;
        this.relType = relType;
        orig = null;
    }

    /**
     * Returns the ID of the link. <br>
     * Link ID gives each TimeML link an identification number
     * Can be null.
     * @since jTLEX 1.0
     * @return the ID of the link.
     */
    public int getLinkID() {
        return linkID;
    }

    /**
     * Returns the type of Link.
     * Will never return null due to constructor.
     * @since jTLEX 1.0
     * @return {@link ITimeMLLink.TLink} / {@link ITimeMLLink.ALink} / {@link ITimeMLLink.SLink}
     */
    public Class<ITimeMLLink> getLinkTag() {
        return this.linktag;
    }

    /**
     * Returns the syntax of the link. <br>
     * Syntax holds any syntactic pattern that was used in generating the link (e.g.
     * from an automatic annotation tool). <br>
     * Can be null.
     * @since jTLEX 1.0
     * @return the syntax of the link.
     */
    public String getSyntax() {
        return syntax;
    }

    /**
     * Returns the ID of the event instance involved in the link.
     * OBLIGATORY, cannot be null
     * @since jTLEX 1.0
     * @return the ID of the event instance involved in the link.
     */
    public String getEventInstance() {
        return eventInstance.getIdStr();
    }

    /**
     * Returns the ID of the relation signal. <br>
     * If the temporal relation holding between the entities
     * is explicitly signaled in the text, then the ID of that signal needs
     * to be supplied here.<br>
     * Can be null.
     * @since jTLEX 1.0
     * @return the optional ID of the relation signal.
     */
    public ITimeMLSignal getSignal() {
        return signal;
    }

    /**
     * Returns whether a TLINK was generated
     * {@code MANUALLY} or by {@code CLOSURE}. <br>
     * THIS IS ONLY FOR TLINK
     * Can be null.
     * @since jTLEX 1.0
     * @return Whether link was generated {@code MANUALLY} or by {@code CLOSURE}
     */
    public Origintype getOrigin() {
        if (!linktag.equals(ITimeMLLink.TLink.class)) {
            String errorMessage = "Error: Origin only used in TLink)";
            throw new IllegalArgumentException(errorMessage);
        }
        return orig;
    }

    /**
     * Returns the ID of the entity that is being related to the event instance
     * with ID=eventInstanceID or time expression with ID=timeID. <br>
     * relatedToEventInstance or relatedToTime needs to be present.
     * Cannot be null due to constructor.
     * @since jTLEX 1.0
     * @return the ID of the entity that is being related to the event instance
     * with ID=eventInstanceID or time expression with ID=timeID.
     */
    public String getRelatedToNode() {
        return relatedToNode.getIdStr();
    }

    /**
     * This is the temporal relation holding between
     * the entities. <br>These are represented by 3 different enums:
     * <ul>
     * <li>ALinkRelType
     * <li>SLinkRelType
     * <li>TLinkRelType
     * </ul>
     * <br>The returned type depends on the linkTag.
     * Cannot be null due to constructor.
     * @since jTLEX 1.0
     * @return {@code ALinkRelType} / {@code SLinkRelType} / {@code TLinkRelType}
     * @see ALink
     * @see SLink
     * @see TLink
     */
    public ITimeMLLinkType getRelType() {
        return relType;
    }


    /**
     * Converts the Link to a String
     * Cannot be null due to requiring items from constructor.
     * @since jTLEX 1.0
     * @return String
     */
    public String toString() {

        String ret = "Link: {" +
                "ID = " + this.getLinkID() +
                ", LinkTag = " + this.getLinkTag().getSimpleName().toUpperCase() +
                ", Syntax = \"" + this.getSyntax() + "\"" +
                ", Temporal Relation = " + this.getRelType();
        if (linktag.equals(ITimeMLLink.TLink.class)) {
            ret += ", Origin = " + this.getOrigin();
        }
        if (this.getSignal() == null) {
            ret += ", Signal = null";
        } else {
            ret += "\n\t" + this.getSignal();
        }
        ret += "\n\tRelated to event - " + this.getRelatedToNode() +
                "\n\tEvent Instance - " + this.getEventInstance() + "\n}";
        return ret;
    }

    /**
     * Returns the JSON (RFC 8259) representation of the Link.
     * Cannot be null due to requiring items from constructor.
     * @since jTLEX 1.0
     * @return the JSON (RFC 8259) representation of the Link.
     */
    public String toJson() {
        String ret = "{\"id\":\"lid" + this.getLinkID() +
                "\", \"linkTag\":\"" + this.getLinkTag().getSimpleName().toUpperCase() +
                "\", \"syntax\":\"" + this.getSyntax() +
                "\", \"temporalRelation\":\"" + this.getRelType() +
                "\"";

        if (linktag.equals(ITimeMLLink.TLink.class) && this.getOrigin() != null) {

            ret += ", \"origin\":\"" + this.getOrigin().toString() + "\"";
        }

        if (this.getSignal() != null)
            ret += ", \"signal\":" + this.getSignal().toJson();
        else
            ret += ", \"signal\":{}";

        ret += ", \"relatedToNode\":\"" + this.getRelatedToNode() +
                "\", \"eventInstance\": \"" + this.getEventInstance() + "\"}";

        return ret;

    }

    /**
     * Returns a field-by-field copy of the Link.
     * Cannot be null due to requiring items from constructor.
     * @since jTLEX 1.0
     * @return a copy of the Link.
     */
    public TimeMLLink clone() {
        TimeMLLink ret = null;
        if (this.getLinkTag().equals(ITimeMLLink.TLink.class)) {
            ret = new TimeMLLink(
                    (Class<?>) this.getLinkTag(),
                    this.getLinkID(),
                    this.getSyntax(),
                    this.eventInstance.clone(),
                    this.getSignal() == null ? null : this.getSignal().clone(),
                    this.relatedToNode.clone(),
                    (ITimeMLLinkType) this.getRelType(),
                    (ITimeMLLink.Origintype) this.getOrigin());
        } else if (this.getLinkTag().equals(ITimeMLLink.ALink.class) ||
                this.getLinkTag().equals(ITimeMLLink.SLink.class)) {
            ret = new TimeMLLink(
                    (Class<?>) this.getLinkTag(),
                    this.getLinkID(),
                    this.getSyntax(),
                    this.eventInstance.clone(),
                    this.getSignal() == null ? null : this.getSignal().clone(),
                    this.relatedToNode.clone(),
                    (ITimeMLLinkType) this.getRelType());
        }
        return ret;
    }
    
    
    /**
	 * Generates a hash code for the link.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
     */
    @Override
    public int hashCode() {
        return (eventInstance.getIdStr() + relatedToNode.getIdStr()).hashCode();
    }
    
	/**
	 * Checks if this link equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both links are equal, i.e. same ID and tag, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLLink))
            return false;

        ITimeMLLink obj = (TimeMLLink) o;

        if (this.linkID != obj.getLinkID()) 
        	return false;
        
        if (this.syntax != obj.getSyntax())
        	return false;

        if (!this.getEventInstance().equals(obj.getEventInstance()))
        	return false;
        
        if (this.signal != null && obj.getSignal() != null) {
        	if (this.signal.getID() != obj.getSignal().getID())
        		return false;
        } else if ((this.signal == null && obj.getSignal() != null) || (this.signal != null && obj.getSignal() == null)) {
        	return false;
        }
        
        if (!this.getRelatedToNode().equals(obj.getRelatedToNode()))
        	return false;
        
        if (this.linktag.isInstance(ITimeMLLink.TLink.class) && obj.getLinkTag().isInstance(ITimeMLLink.TLink.class)) {
        	if (this.orig != obj.getOrigin())
        		return false;
        }
        
        if (this.linktag != obj.getLinkTag())
        	return false;
        
        if (this.relType != obj.getRelType())
        	return false;
               
        return true;
    }
}
