package edu.fiu.jtlex.timeml.data;

import java.util.Objects;

/**
 * Default implementation of {@code ISignal}.
 * @since jTLEX 1.0
 * @author arada002
 *
 */
public class TimeMLSignal implements ITimeMLSignal {

	private final int signalID;
	private final String signalString;
	
	/**
	 * Creates a new Signal object.
	 * @param signalID						a unique ID number, must be an integer or will be undefined
	 * @param signalString					temporal preposition, conjunction or modifier phrase
	 * @throws IllegalArgumentException 	if signalID is less than 1
	 * 										if signalString is empty or all whitespace
	 * @throws NullPointerException 		if signalString is null
	 */
	public TimeMLSignal(int signalID, String signalString) {
		if (signalID < 1) {
			throw new IllegalArgumentException("Signal ID may not be less than 1");
		}
		if (signalString == null) {
			throw new NullPointerException("Signal string can not be null");
		}
		signalString = signalString.trim();
		if (signalString.length() == 0) {
			throw new IllegalArgumentException("Signal string can not be empty string or "
					+ "all whitespace");
		}
		this.signalID = signalID;
		this.signalString = signalString;
	}

	/**
	 * Returns the ID of a Signal.
	 * Will never return null due to it being required in constructor.
	 * @since jTLEX 1.0
	 * @return the ID of a Signal. 
	 */
	public int getID() {
		return signalID;
	}
	
	/**
	 * Returns the String format of a Signal ID.
	 * Will never return null since it is required in constructor.
	 * @since jTLEX 1.0
	 * @return the ID of a Signal in String format. 
	 */
	public String getIdStr() {
		return "sid"+signalID;
	}
	
	/**
	 * Returns the String of a Signal.
	 * Will never return null because constructor requires a valid string.
	 * @since jTLEX 1.0
	 * @return the String of a Signal. 
	 */
	public String getSignalString() {
		return signalString;
	}
	
	/**
	 * Returns the Signal ID and String in String format.
	 * Will never return null due to constructor requiring them.
	 * @since jTLEX 1.0
	 * @return the Signal ID and String as a String.
	 */
	@Override
	public String toString() {
		return "Signal: {Id = " + this.getIdStr() + ", String = \"" + this.getSignalString() + "\"}";
	}
	
	/**
	 * Returns the String ID and String in JSON format.
	 * Will never return null due to constructor.
	 * @since jTLEX 1.0
	 * @return JSON Formatted String of Signal ID and String. 
	 */
	public String toJson() {
		return "{\"id\":\"" + this.getIdStr() + "\", \"signalString\":\"" + this.getSignalString() + "\"}";
	}

	/**
	 * Returns a field-by-field copy clone of the Signal.
	 * Will never return null, will always return a valid clone.
	 * @since jTLEX 1.0
	 * @return copy of Signal. 
	 */
	public TimeMLSignal clone() {
		return new TimeMLSignal(this.signalID, this.signalString);
	}
	
	/**
	 * Generates a hash code for the signal.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode() {
        int hashCode = 1;
        hashCode = 31*hashCode + this.signalID;
        hashCode = 31*hashCode + (this.signalString==null ? 0 : this.signalString.hashCode());
        
        return hashCode;
    }
	/**
	 * Checks if this Signal equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both signals are equal, i.e. same ID and string, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLSignal))
            return false;

        ITimeMLSignal obj = (TimeMLSignal) o;

        if (this.signalID != obj.getID())
            return false;
        
        if (this.signalString != obj.getSignalString())
        	return false;

        return true;
    }

}
