package edu.fiu.jtlex.timeml.graph;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import edu.fiu.jtlex.timeml.data.ITimeMLEvent;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance;
import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.ITimeMLLink.SLink;
import edu.fiu.jtlex.timeml.data.ITimeMLText;
import edu.fiu.jtlex.timeml.io.ITimeMLParser;
import edu.fiu.jtlex.timeml.io.TimeMLParser;

/**
 * Represents a TimeML Graph, and the default implementation of {@link ITimeMLGraph}.
 *
 * A TimeML temporal graph is a graph T = (V,E) where: V is a set of temporal events and
 * time expressions. E is a set of TimeML links: a tuple (u,v,w) where u, v belong to V
 * and w is a link type.
 *
 * There are 25 link types, refer to {@link edu.fiu.jtlex.timeml.data.ITimeMLLink.TLink TLink},
 * {@link edu.fiu.jtlex.timeml.data.ITimeMLLink.ALink ALink}, and {@link edu.fiu.jtlex.timeml.data.ITimeMLLink.SLink SLink}.
 *
 * A graph is represented with two Hash Tables, one for Nodes and one for Links.
 *
 * @since jTLEX 1.0
 */
public class TimeMLGraph implements ITimeMLGraph {

    private Map<String, ITimeMLNode> nodes = new HashMap<>();
    private Map<Integer, ITimeMLLink> links = new HashMap<>();
    private Map<String, ITimeMLEvent> events = new HashMap<>();
    private GraphType type = GraphType.MainGraph;
    private Map<String, Set<ITimeMLLink>> inLinks = new HashMap<>();
    private Map<String, Set<ITimeMLLink>> outLinks = new HashMap<>();
    

    /**
     * Creates an empty graph.
     */
    public TimeMLGraph() {}

	/**
	 * Utilizes an {@link ITimeMLParser} to parse a stream into
     * nodes and edges, and adds them to the graph.
	 * 
	 * @param stream The InputStream to parse
     * @throws NullPointerException If the stream is null
     * @since jTLEX 1.0
	 */
    public TimeMLGraph(InputStream stream) {
        Objects.requireNonNull(stream);

        ITimeMLParser parser = new TimeMLParser();
        ITimeMLText text;
        try {
            text = parser.parse(stream);
        } catch (UnsupportedEncodingException e) {
            return;
        }

        text.getTimexes().forEach(this::addNode);
        text.getInstances().forEach(this::addNode);
        text.getLinks().forEach(this::addLink);
        text.getEvents().forEach(this::addEvent);

    }

	/**
	 * Adds the provided link to the graph. The endpoints of the link must already
	 * be present in the Graph.
	 *
	 * @param link The ILink to be added.
	 * @return True if the link was added, otherwise it returns false if the link
	 *         already exists or if the nodes were not in the graph, and as a
	 *         result, the link is not added.
	 * @throws NullPointerException     if the link is null.
	 * @throws IllegalArgumentException if nodes are not in graph before link is
	 *                                  added.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean addLink(ITimeMLLink link) throws IllegalArgumentException {
        Objects.requireNonNull(link);

        int linkId = link.getLinkID();
        if (links.containsKey(linkId))
            return false;
        
        Set<ITimeMLNode> checkNodes = this.getNodes();
        boolean hasEventNode = false;
        boolean hasRelatedToNode = false;
        
        for (ITimeMLNode checkNode : checkNodes) {
        	if (checkNode.getIdStr().equals(link.getEventInstance())) {
        			hasEventNode = true;
        	}
        	
        	if (checkNode.getIdStr().equals(link.getRelatedToNode())) {
        			hasRelatedToNode = true;
        	}
        	
        }
        
        if (!hasEventNode || !hasRelatedToNode) {
        	throw new IllegalArgumentException("Nodes must be inside Graph first before Link is added");
        }

        Set<ITimeMLLink> newOutLinks = new HashSet<ITimeMLLink>();
    	newOutLinks = outLinks.get(link.getEventInstance());
    	if (newOutLinks == null) {
    		newOutLinks = new HashSet<ITimeMLLink>();
    		newOutLinks.add(link);
    	} else {
    	newOutLinks.add(link);
    	}
    	outLinks.put(link.getEventInstance(), newOutLinks);
    	
        Set<ITimeMLLink> newInLinks = new HashSet<ITimeMLLink>();
    	newInLinks = inLinks.get(link.getRelatedToNode());
    	if (newInLinks == null) {
    		newInLinks = new HashSet<ITimeMLLink>();
    		newInLinks.add(link);
    	} else {
    		newInLinks.add(link);
    	}
    	inLinks.put(link.getRelatedToNode(), newInLinks);

        links.put(link.getLinkID(), link);

        return true;
    }


	/**
	 * Gets a node by id.
	 * 
	 * @param nodeId The id of the node to retrieve.
	 * @return The node with the given id, or null if no such node exists.
	 * @since jTLEX 1.0
	 */
    @Override
    public ITimeMLNode getNodeById(String nodeId) {
        return nodes.get(nodeId);
    }
    
	/**
	 * Gets a link by id.
	 *
	 * @param linkId The id of the link to retrieve.
	 * @return The link with the given id, or null if no such link exists.
	 * @since jTLEX 1.0
	 */
    @Override
    public ITimeMLLink getLinkById(int linkId) {
        return links.get(linkId);
    }
    
    /**
	 * Gets a event by id.
	 * 
	 * @param eventId The id of the event to retrieve.
	 * @return The event with the given id, or null if no such event exists.
	 * @since jTLEX 1.0
	 */
    @Override
    public ITimeMLEvent getEventById(String eventId) {
        return events.get(eventId);
    }

	/**
	 * Removes a provided link from the graph.
	 *
	 * @param link The ILink to remove.
	 * @return true if link is removed, false if not removed or if link was not found.
     * @throws NullPointerException If ILink is null
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeLink(ITimeMLLink link) {
        Objects.requireNonNull(link);
        return removeLinkById(link.getLinkID());
    }

    
	/**
	 * Removes a link by id.
	 * 
	 * @param linkId The id of the link to remove.
	 * @return true if the link is removed, false if not removed or if link was not found.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeLinkById(int linkId) {
    	
        ITimeMLLink link = this.getLinkById(linkId);
        if (link == null)
        	return false;

        String source = link.getEventInstance();
        String destination = link.getRelatedToNode();
        ITimeMLLink removedLink = links.remove(link.getLinkID());
        if (removedLink == null)
            return false;

        Set<ITimeMLLink> toRemoveOutLinks = outLinks.get(source);
        toRemoveOutLinks.remove(link);
    	outLinks.put(source, toRemoveOutLinks);

        
        Set<ITimeMLLink> toRemoveInLinks = inLinks.get(destination);
        toRemoveInLinks.remove(link);
    	inLinks.put(destination, toRemoveInLinks);
    	
    	boolean removed = true;
    	
    	for (ITimeMLLink links : this.getLinks()) {
    		if (links == link) {
    			removed = false;
    		}
    	}

        return removed;
    }
    
    
	/**
	 * Adds a node to the graph.
	 *
	 * @param node The node to add
	 * @return True if the node was added, or False if the node already exists.
     * @throws NullPointerException If the node is null
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean addNode(ITimeMLNode node) {
        Objects.requireNonNull(node);
        String nodeId = node.getIdStr();
        if (nodes.containsKey(nodeId))
            return false;
        nodes.put(nodeId, node);
        
        Set<ITimeMLLink> emptyOutLinks = new HashSet<ITimeMLLink>();
        outLinks.put(node.getIdStr(), emptyOutLinks);
        
        Set<ITimeMLLink> emptyInLinks = new HashSet<ITimeMLLink>();
        inLinks.put(node.getIdStr(), emptyInLinks);
        return true;
    }

	/**
	 * Removes a provided node from the graph. Removes links it is involved with as
	 * well.
	 *
	 * @param node The node to remove.
	 * @return True if removed, false if not removed or null input.
	 * @throws NullPointerException if the node is null.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeNode(ITimeMLNode node) {
        Objects.requireNonNull(node);
        return removeNodeById(node.getIdStr());
    }

	/**
	 * Removes a node by id. Removes links it is involved with as well.
	 *
	 * @param nodeId The id of the node to remove.
	 * @return True if removed, false if not removed.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeNodeById(String nodeId) {
    	ITimeMLNode node = this.nodes.remove(nodeId);
    	if (node == null)
    		return false;
    	
    	for (ITimeMLLink link : this.getLinks()) {
    		if (link.getEventInstance().equals(node.getIdStr()))
    			this.removeLink(link);
    		if (link.getRelatedToNode().equals(node.getIdStr()))
    			this.removeLink(link);
    	}
    	boolean removed = true;
    	for (ITimeMLNode nodes : this.getNodes()) {
    		if (nodeId == nodes.getIdStr())
    			removed = false;
    	}
       return removed;
    }
    
    
	/**
	 * Adds an event to the graph.
	 *
	 * @param event The event to add
	 * @return True if the event was added, or False if the event already exists.
     * @throws NullPointerException If the node is null
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean addEvent(ITimeMLEvent event) {
        Objects.requireNonNull(event);
        String eventId = event.getIdStr();
        if (events.containsKey(eventId))
            return false;
        events.put(eventId, event);
		return true;
        
    }

	/**
	 * Removes a provided event from the graph. Removes instances it is involved with as
	 * well.
	 *
	 * @param event The event to remove.
	 * @return True if removed, false if not removed or null input.
	 * @throws NullPointerException if the node is null.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeEvent(ITimeMLEvent event) {
        Objects.requireNonNull(event);
        return removeEventById(event.getIdStr());
    }

	/**
	 * Removes an event by id. Removes instances it is involved with as well.
	 *
	 * @param eventId The id of the event to remove.
	 * @return True if removed, false if not removed, null, or if is a timex.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean removeEventById(String eventId) {
    	ITimeMLEvent event = this.events.remove(eventId);
    	if (event == null)
    		return false;

    	boolean removedNode = false;
 		for (ITimeMLNode node : this.getNodes()) {

 			if (node instanceof ITimeMLInstance) {
 		    	ITimeMLInstance instance = (ITimeMLInstance) node;
 	       		if (eventId.equals(instance.getEvent())) {

 	       			removeNodeById(node.getIdStr()); 				
 	       			removedNode = true;
 	       		}
 			}
       	}
    	boolean removed = true;
    	if (removedNode && removed)
    		return true;
    	else
    		return false;
    }

	/**
	 * Returns a snapshot of all the nodes currently in the graph. Modifying this
	 * set won't modify the actual graph's node set.
	 * 
	 * @return A set containing the nodes currently in the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public Set<ITimeMLNode> getNodes() {
        return new HashSet<ITimeMLNode>(nodes.values());
    }
    
	/**
	 * Returns a snapshot of all the links currently in the graph. Modifying this
	 * set won't modify the actual graph's link set.
	 * 
	 * @return A set containing the links currently in the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public Set<ITimeMLLink> getLinks() {
        return new HashSet<ITimeMLLink>(links.values());
    }
    
	/**
	 * Returns a snapshot of all the events currently in the graph. Modifying this
	 * set won't modify the actual graph's event set.
	 * 
	 * @return A set containing the events currently in the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public Set<ITimeMLEvent> getEvents() {
        return new HashSet<ITimeMLEvent>(events.values());
    }
    
	/**
	 * Returns the OutLinks in the graph.
	 *
	 * @return A map with the OutLinks in the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public Map<String, Set<ITimeMLLink>> getOutLinks() {
        return this.outLinks;
    }
    
    
	/**
	 * Returns the InLinks in the graph.
	 *
	 * @return A map of the InLinks in the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public Map<String, Set<ITimeMLLink>> getInLinks() {
        return this.inLinks;
    }
    
    
	/**
	 * Gets the Graph's type.
	 *
	 * @return A {@link  GraphType} representing the graph type.
	 * @since jTLEX 1.0
	 */
    @Override
    public GraphType getGraphType() {
        return this.type;
    }
    
	/**
	 * Gets the Graph's Parent Graph. Used for obtaining the parent graph from
	 * Subordination Graphs.
	 *
	 * @return A TimeMLGraph representing the parent, or null if it is not a
	 *         subordination graph, meaning there is no parent.
	 * @since jTLEX 1.0
	 */
	@Override
	public ITimeMLGraph getParent() {
		return null;
	}
    
	/**
	 * Creates a sub graph.
	 * 
	 * @param nodes	The set of {@link ITimeMLNode}.
	 * @param links	The set of {@link ITimeMLLink}.
	 * @return an ITimeMLGraph that is the subgraph.
	 * @throws IllegalArgumentException if the nodes or links are not inside the graph
	 * the subgraph is being created from 
	 * @since jTLEX 1.0
	 */
    @Override
    public ITimeMLGraph subGraph(Set<ITimeMLNode> nodes, Set<ITimeMLLink> links, Set<ITimeMLEvent> events) throws IllegalArgumentException {
    	for (ITimeMLNode node : nodes) {
    		if (!this.getNodes().contains(node)) {
    			throw new IllegalArgumentException("Invalid node set");
    		}
    	}
    	
    	for (ITimeMLLink link : links) {
    		if (!this.getLinks().contains(link)) {
    			throw new IllegalArgumentException("Invalid link set");
    		}
    	}
    	
    	for (ITimeMLEvent event : events) {
    		if (!this.getEvents().contains(event)) {
    			throw new IllegalArgumentException("Invalid event set");
    		}
    	}
    	ITimeMLGraph subGraph = new TimeMLSubGraph(nodes, links, events, this);
    	return subGraph;
    	
    }
	
	/**
	 * Creates a clone of the graph by first cloning the nodes and links and then returning 
	 * them inside a new Graph object.
	 *
	 * @return A copy of the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public ITimeMLGraph clone() {
        ITimeMLGraph clone = new TimeMLGraph();
        for (ITimeMLNode node : this.nodes.values()) {
            clone.addNode(node.clone());
        }
        for (ITimeMLLink link : this.links.values()) {
            clone.addLink(link.clone());
        }
        return clone;
    }
    
	/**
	 * Returns the JSON (RFC 8259) representation of the graph.
	 *
	 * @return A flattened JSON string of the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public String toJson() {
        StringBuilder sb = new StringBuilder(1024);

        sb.append("{");

        sb.append("\"nodeIds\": ").append("[");
        for (ITimeMLNode node : getNodes()) {
            sb.append("\"").append(node.getIdStr()).append("\", ");
        }
        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" ") + 1);
        sb.append("], ");

        sb.append("\"links\": ").append("[");
        for (ITimeMLLink link : getLinks()) {
            sb.append("{");
            sb.append("\"id\": \"lid").append(link.getLinkID()).append("\", ");
            sb.append("\"from\": \"").append(link.getEventInstance()).append("\", ");
            sb.append("\"type\": \"").append(link.getRelType()).append("\", ");
            sb.append("\"to\": \"").append(link.getRelatedToNode()).append("\"");
            sb.append("}, ");
        }
        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" ") + 1);
        sb.append("]");

        sb.append("}");

        return sb.toString();
    }

	/**
	 * Gets a printable string representation of the graph.
	 *
	 * @return A printable string representation of the graph.
	 * @since jTLEX 1.0
	 */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(128);

        sb.append("Graph Type: ").append(getGraphType().toString()).append("\n");
        sb.append("Nodes Count = ").append(getNodes().size()).append("\n");
        sb.append("Links count = ").append(getLinks().size()).append("\n");

        for (Map.Entry<Class<?>, Integer> tuple : ITimeMLGraph.getLinkTypesCount(this).entrySet()) {
            sb.append(tuple.getKey()).append(": ").append(tuple.getValue()).append("\n");
        }

        sb.append("Nodes:").append("\n");
        int nodeCount = 0;
        for (String nodeId : nodes.keySet()) {
            nodeCount += 1;
            ITimeMLNode node = getNodeById(nodeId);
            sb.append(node.getIdStr());
            sb.append(nodeCount % 10 == 0 ? "\n" : ", ");
        }
        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" "));
        sb.append("\n");

        sb.append("Links: (From -> To)").append("\n");
        for (int linkId : links.keySet()) {
            ITimeMLLink link = getLinkById(linkId);
            sb.append("(").append(link.getEventInstance())
                    .append(" ").append(link.getRelType())
                    .append(" ").append(link.getRelatedToNode())
                    .append(")").append("\n");
        }

        return sb.toString();
    }
    
    
	/**
	 * Generates a hash code for the graph, based of its nodes and edges.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode() {
        int hashCode = 1;
        hashCode = 31*hashCode + (int)(this.nodes ==null ? 0 : this.nodes.hashCode());
        hashCode = 31*hashCode + (int)(this.links==null ? 0 : this.links.hashCode());
        return hashCode;
    }

	/**
	 * Checks if this graph equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both graphs are equal, i.e. same set of nodes and links, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o) {
        Objects.requireNonNull(o);

        if (!(o instanceof ITimeMLGraph))
            return false;

        ITimeMLGraph obj = (TimeMLGraph) o;

        if (this.getNodes().size() != obj.getNodes().size())
            return false;

        for (String nodeId : this.nodes.keySet()) {
            if (obj.getNodeById(nodeId) == null)
                return false;
        }

        if (this.getLinks().size() != obj.getLinks().size())
            return false;

        for (int linkId : this.links.keySet()) {
            if (obj.getLinkById(linkId) == null)
                return false;
        }

        return true;
    }

	/**
	 * Represents a TimeML SubGraph.
	 *
	 * A TimeML temporal subgraph is a graph that can be considered a view of a part
	 * of a graph and cannot contain any S-Links.
	 *
	 * @since jTLEX 1.0
	 */
	private static class TimeMLSubGraph implements ITimeMLGraph {
		
	    private Map<String, ITimeMLNode> subNodes = new HashMap<>();
	    private Map<Integer, ITimeMLLink> subLinks = new HashMap<>();
	    private Map<String, ITimeMLEvent> subEvents = new HashMap<>();
	    private Map<String, Set<ITimeMLLink>> subInLinks = new HashMap<>();
	    private Map<String, Set<ITimeMLLink>> subOutLinks = new HashMap<>();
		private GraphType type = GraphType.SubordinationGraph;
		private ITimeMLGraph parent;
		
		/**
		 * Creates a subgraph based on nodes and links and the parent graph.
		 * 
		 * @param nodes	The set of {@link ITimeMLNode}.
		 * @param links	The set of {@link ITimeMLLink}.
		 * @param parent the parent {@link ITimeMLGraph}.
		 * @return a created SubGraph Object
		 * @throws IllegalArgumentException if there is an attempt to add an S-Link to the Sub-Graph
	     * @since jTLEX 1.0
		 */
		public TimeMLSubGraph(Set<ITimeMLNode> nodes, Set<ITimeMLLink> links, Set<ITimeMLEvent> events, ITimeMLGraph parent) throws IllegalArgumentException {
			this.parent = parent;
	        for (ITimeMLLink L : links) {
	            if (L.getRelType() instanceof SLink) {
	    			throw new IllegalArgumentException("Cannot contain an S-Link in subgraph.");
	           }
	       }
	        
	        for (ITimeMLNode node : nodes)
	    		this.addNode(node);
	    	for (ITimeMLLink link : links)
	    		this.addLink(link);
	    	for (ITimeMLEvent event : events)
	    		this.addEvent(event);
		}
		    
		/**
		 * Adds the provided link to both the subgraph and the parent graph. The
		 * endpoints of the link must already be present in the Graph.
		 *
		 * @param link The ILink to be added.
		 * @return True if the link was added, otherwise it returns false if the link
		 *         already exists or if the nodes were not in the graph, and as a
		 *         result, the link is not added.
		 * @throws NullPointerException     if the link is null.
		 * @throws IllegalArgumentException if the link is an S-Link or if the nodes are
		 *                                  not in graph before link is added.
		 * @since jTLEX 1.0
		 */
		@Override
		public boolean addLink(ITimeMLLink link) throws IllegalArgumentException {
	            if (link.getRelType() instanceof SLink) {
	    			throw new IllegalArgumentException("Cannot contain an S-Link in subgraph.");
	           }
			parent.addLink(link);
			
			
			Objects.requireNonNull(link);

	        int linkId = link.getLinkID();
	        if (subLinks.containsKey(linkId))
	            return false;
	        
	        Set<ITimeMLNode> checkNodes = this.getNodes();
	        boolean hasEventNode = false;
	        boolean hasRelatedToNode = false;
	        for (ITimeMLNode checkNode : checkNodes) {
	        	if (checkNode.getIdStr().equals(link.getEventInstance())) {
	        			hasEventNode = true;
	        	}
	        	
	        	if (checkNode.getIdStr().equals(link.getRelatedToNode())) {
	        			hasRelatedToNode = true;
	        	}
	        	
	        }
	        
	        if (!hasEventNode || !hasRelatedToNode) {
	        	throw new IllegalArgumentException("Nodes must be inside Graph first before Link is added");
	        }

	        Set<ITimeMLLink> newOutLinks = new HashSet<ITimeMLLink>();
	    	newOutLinks = subOutLinks.get(link.getEventInstance());
	    	if (newOutLinks == null) {
	    		newOutLinks = new HashSet<ITimeMLLink>();
	    		newOutLinks.add(link);
	    	} else {
	    	newOutLinks.add(link);
	    	}
	    	subOutLinks.put(link.getEventInstance(), newOutLinks);
	    	
	        Set<ITimeMLLink> newInLinks = new HashSet<ITimeMLLink>();
	    	newInLinks = subInLinks.get(link.getRelatedToNode());
	    	if (newInLinks == null) {
	    		newInLinks = new HashSet<ITimeMLLink>();
	    		newInLinks.add(link);
	    	} else {
	    		newInLinks.add(link);
	    	}
	    	subInLinks.put(link.getRelatedToNode(), newInLinks);

	        subLinks.put(link.getLinkID(), link);

	        return true;
			}
		
		/**
		 * Adds a node to the subgraph and parent graph.
		 *
		 * @param node The node to add
		 * @return True if the node was added, or False if the node already exists.
	     * @throws NullPointerException If the node is null
		 * @since jTLEX 1.0
		 */
		@Override
		 public boolean addNode(ITimeMLNode node) {
			 	parent.addNode(node);
				
			 	Objects.requireNonNull(node);
		        String nodeId = node.getIdStr();
		        if (subNodes.containsKey(nodeId))
		            return false;
		        subNodes.put(nodeId, node);
		        
		        Set<ITimeMLLink> emptyOutLinks = new HashSet<ITimeMLLink>();
		        subOutLinks.put(node.getIdStr(), emptyOutLinks);
		        
		        Set<ITimeMLLink> emptyInLinks = new HashSet<ITimeMLLink>();
		        subInLinks.put(node.getIdStr(), emptyInLinks);
		        return true;
				 
		 }
		
		/**
		 * Removes a provided node from both the subgraph and it's parent Graph. Removes links it is involved with as
		 * well.
		 *
		 * @param node The node to remove.
		 * @return True if removed, false if not removed or null input.
		 * @throws NullPointerException if the node is null.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public boolean removeNode(ITimeMLNode node) {
	        Objects.requireNonNull(node);

	        parent.removeNode(node);
	        return removeNodeById(node.getIdStr());
	    }

		/**
		 * Gets a node by id.
		 * 
		 * @param nodeId The id of the node to retrieve.
		 * @return The node with the given id, or null if no such node exist.
		 * @since jTLEX 1.0
		 */
		@Override
		public ITimeMLNode getNodeById(String nodeId) {
	        return subNodes.get(nodeId);
		}

		/**
		 * Gets a link by id.
		 *
		 * @param linkId The id of the link to retrieve.
		 * @return The link with the given id, or null if no such link exist.
		 * @since jTLEX 1.0
		 */
		@Override
		public ITimeMLLink getLinkById(int linkId) {
	        return subLinks.get(linkId);
		}
		
		/**
		 * Gets a event by id.
		 * 
		 * @param eventId The id of the event to retrieve.
		 * @return The event with the given id, or null if no such event exists.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public ITimeMLEvent getEventById(String eventId) {
	        return subEvents.get(eventId);
	    }

		/**
		 * Removes a provided link from both the subgraph and the parent graph.
		 *
		 * @param link The ILink to remove.
		 * @return true if link is removed, false if not removed or if link was not
		 *         found.
		 * @throws NullPointerException If ILink is null
		 * @since jTLEX 1.0
		 */
		@Override
		public boolean removeLink(ITimeMLLink link) {
	        Objects.requireNonNull(link);

			parent.removeLink(link);
	        return removeLinkById(link.getLinkID());
		}

		/**
		 * Removes a node by id from both the subgraph and the parent graph. Removes
		 * links it is involved with as well.
		 *
		 * @param nodeId The id of the node to remove.
		 * @return True if removed, false if not removed.
		 * @since jTLEX 1.0
		 */
		@Override
		public boolean removeNodeById(String nodeId) {
			parent.removeNodeById(nodeId);
			
			ITimeMLNode node = this.subNodes.remove(nodeId);
	    	if (node == null)
	    		return false;
	    	
	    	for (ITimeMLLink link : this.getLinks()) {
	    		if (link.getEventInstance().equals(node.getIdStr()))
	    			this.removeLink(link);
	    		if (link.getRelatedToNode().equals(node.getIdStr()))
	    			this.removeLink(link);
	    	}
	    	boolean removed = true;
	    	for (ITimeMLNode nodes : this.getNodes()) {
	    		if (nodeId.equals(nodes.getIdStr()))
	    			removed = false;
	    	}
	       return removed;
		}

		/**
		 * Removes a link by id from both the subgraph and the parent graph.
		 * 
		 * @param linkId The id of the link to remove.
		 * @return true if the link is removed, false if not removed or if link was not
		 *         found.
		 * @since jTLEX 1.0
		 */
		@Override
		public boolean removeLinkById(int linkId) {
			parent.removeLinkById(linkId);
			 
		        ITimeMLLink link = this.getLinkById(linkId);
		        if (link == null)
		        	return false;

		        String source = link.getEventInstance();
		        String destination = link.getRelatedToNode();
		        ITimeMLLink removedLink = subLinks.remove(link.getLinkID());
		        if (removedLink == null)
		            return false;

		        Set<ITimeMLLink> toRemoveOutLinks = subOutLinks.get(source);
		        toRemoveOutLinks.remove(link);
		    	subOutLinks.put(source, toRemoveOutLinks);

		        
		        Set<ITimeMLLink> toRemoveInLinks = subInLinks.get(destination);
		        toRemoveInLinks.remove(link);
		    	subInLinks.put(destination, toRemoveInLinks);
		    	
		    	boolean removed = true;
		    	
		    	for (ITimeMLLink links : this.getLinks()) {
		    		if (links == link) {
		    			removed = false;
		    		}
		    	}

		        return removed;
		}

		
		/**
		 * Adds an event to the graph.
		 *
		 * @param event The event to add
		 * @return True if the event was added, or False if the event already exists.
	     * @throws NullPointerException If the node is null
		 * @since jTLEX 1.0
		 */
	    @Override
	    public boolean addEvent(ITimeMLEvent event) {
	        Objects.requireNonNull(event);
		 	parent.addEvent(event);

	        String eventId = event.getIdStr();
	        if (subEvents.containsKey(eventId))
	            return false;
	        subEvents.put(eventId, event);
			return true;
	        
	    }

		/**
		 * Removes a provided event from the graph. Removes instances it is involved with as
		 * well.
		 *
		 * @param event The event to remove.
		 * @return True if removed, false if not removed or null input.
		 * @throws NullPointerException if the node is null.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public boolean removeEvent(ITimeMLEvent event) {
	        Objects.requireNonNull(event);
	        return removeEventById(event.getIdStr());
	    }

		/**
		 * Removes an event by id. Removes instances it is involved with as well.
		 *
		 * @param eventId The id of the event to remove.
		 * @return True if removed, false if not removed, null, or if is a timex.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public boolean removeEventById(String eventId) {
	    	ITimeMLEvent event = this.subEvents.remove(eventId);
	    	if (event == null)
	    		return false;

			parent.removeEventById(eventId);
	    	boolean removedNode = false;
	 		for (ITimeMLNode node : this.getNodes()) {

	 			if (node instanceof ITimeMLInstance) {
	 		    	ITimeMLInstance instance = (ITimeMLInstance) node;
	 	       		if (eventId.equals(instance.getEvent()))
	 	       			removeNodeById(node.getIdStr()); 				
	 	       			removedNode = true;
	 			}
	       	}
	    	boolean removed = true;
	    	if (removedNode && removed)
	    		return true;
	    	else
	    		return false;
	    }
	    
		/**
		 * Returns a snapshot of all the nodes currently in the subgraph. Modifying this
		 * set won't modify the actual subgraph's node set.
		 * 
		 * @return A set containing the nodes currently in the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public Set<ITimeMLNode> getNodes() {
			return new HashSet<ITimeMLNode>(subNodes.values());
		}

		/**
		 * Returns a snapshot of all the links currently in the subgraph. Modifying this
		 * set won't modify the actual subgraph's link set.
		 * 
		 * @return A set containing the links currently in the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public Set<ITimeMLLink> getLinks() {
	        return new HashSet<ITimeMLLink>(subLinks.values());
		}

		/**
		 * Returns a snapshot of all the events currently in the graph. Modifying this
		 * set won't modify the actual graph's event set.
		 * 
		 * @return A set containing the events currently in the graph.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public Set<ITimeMLEvent> getEvents() {
	        return new HashSet<ITimeMLEvent>(subEvents.values());
	    }
	    
		/**
		 * Returns the OutLinks in the subgraph.
		 *
		 * @return A map with the OutLinks in the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public Map<String, Set<ITimeMLLink>> getOutLinks() {
			return this.subOutLinks;
		}

		/**
		 * Returns the InLinks in the subgraph.
		 *
		 * @return A map of the InLinks in the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public Map<String, Set<ITimeMLLink>> getInLinks() {
			return this.subInLinks;
		}

		/**
		 * Gets the SubGraph's type. Will always be a Subordination Graph.
		 *
		 * @return A {@link  GraphType} representing the graph type.
		 * @since jTLEX 1.0
		 */
		@Override
		public GraphType getGraphType() {
			return this.type;
		}
		
		/**
		 * Gets the Graph's Parent Graph. Used for obtaining the parent graph from
		 * Subordination Graphs.
		 *
		 * @return A TimeMLGraph representing the parent.
		 * @since jTLEX 1.0
		 */
		@Override
		public ITimeMLGraph getParent() {
			return this.parent;
		}
		
		/**
		 * Creates a clone of the subgraph by first cloning the nodes and links and then
		 * returning them inside a new SubGraph object.
		 *
		 * @return A copy of the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public ITimeMLGraph clone() {
			TimeMLSubGraph clone = new TimeMLSubGraph(this.getNodes(), this.getLinks(), this.getEvents(), this.parent);
			return clone;
		}
	    
		/**
		 * Returns the JSON (RFC 8259) representation of the subgraph.
		 *
		 * @return A flattened JSON string of the subgraph.
		 * @since jTLEX 1.0
		 */
		@Override
		public String toJson() {
			StringBuilder sb = new StringBuilder(1024);

	        sb.append("{");

	        sb.append("\"nodeIds\": ").append("[");
	        for (ITimeMLNode node : this.getNodes()) {
	            sb.append("\"").append(node.getIdStr()).append("\", ");
	        }
	        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" ") + 1);
	        sb.append("], ");

	        sb.append("\"links\": ").append("[");
	        for (ITimeMLLink link : this.getLinks()) {
	            sb.append("{");
	            sb.append("\"id\": \"lid").append(link.getLinkID()).append("\", ");
	            sb.append("\"from\": \"").append(link.getEventInstance()).append("\", ");
	            sb.append("\"type\": \"").append(link.getRelType()).append("\", ");
	            sb.append("\"to\": \"").append(link.getRelatedToNode()).append("\"");
	            sb.append("}, ");
	        }
	        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" ") + 1);
	        sb.append("]");

	        sb.append("}");

	        return sb.toString();
		}

		/**
		 * Creates a sub graph.
		 * 
		 * @param nodes	The set of {@link ITimeMLNode}.
		 * @param links	The set of {@link ITimeMLLink}.
		 * @return an ITimeMLGraph that is the subgraph.
		 * @throws IllegalArgumentException if the nodes or links are not inside the graph
		 * the subgraph is being created from 
		 * @since jTLEX 1.0
		 */
		@Override
		public ITimeMLGraph subGraph(Set<ITimeMLNode> nodes, Set<ITimeMLLink> links, Set<ITimeMLEvent> events) throws IllegalArgumentException {
	    	for (ITimeMLNode node : nodes) {
	    		if (!this.getNodes().contains(node)) {
	    			throw new IllegalArgumentException("Invalid node set");
	    		}
	    	}
	    	
	    	for (ITimeMLLink link : links) {
	    		if (!this.getLinks().contains(link)) {
	    			throw new IllegalArgumentException("Invalid link set");
	    		}
	    	}
	    	ITimeMLGraph subGraph = new TimeMLSubGraph(nodes, links, events, this);
	    	return subGraph;
		}
		
		/**
		 * Gets a printable string representation of the graph.
		 *
		 * @return A printable string representation of the graph.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public String toString() {
	        StringBuilder sb = new StringBuilder(128);

	        sb.append("Graph Type: ").append(getGraphType().toString()).append("\n");
	        sb.append("Nodes Count = ").append(getNodes().size()).append("\n");
	        sb.append("Links count = ").append(getLinks().size()).append("\n");

	        for (Map.Entry<Class<?>, Integer> tuple : ITimeMLGraph.getLinkTypesCount(this).entrySet()) {
	            sb.append(tuple.getKey()).append(": ").append(tuple.getValue()).append("\n");
	        }

	        sb.append("Nodes:").append("\n");
	        int nodeCount = 0;
	        for (String nodeId : this.subNodes.keySet()) {
	            nodeCount += 1;
	            ITimeMLNode node = getNodeById(nodeId);
	            sb.append(node.getIdStr());
	            sb.append(nodeCount % 10 == 0 ? "\n" : ", ");
	        }
	        sb.delete(sb.lastIndexOf(","), sb.lastIndexOf(" "));
	        sb.append("\n");

	        sb.append("Links: (From -> To)").append("\n");
	        for (int linkId : this.subLinks.keySet()) {
	            ITimeMLLink link = getLinkById(linkId);
	            sb.append("(").append(link.getEventInstance())
	                    .append(" ").append(link.getRelType())
	                    .append(" ").append(link.getRelatedToNode())
	                    .append(")").append("\n");
	        }

	        return sb.toString();
	    }
	    
	    
		/**
		 * Generates a hash code for the graph, based of its nodes and edges.
		 *
		 * @return The hash code.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public int hashCode() {
	        int hashCode = 1;
	        hashCode = 31*hashCode + (int)(this.subNodes ==null ? 0 : this.subNodes.hashCode());
	        hashCode = 31*hashCode + (int)(this.subLinks==null ? 0 : this.subLinks.hashCode());
	        return hashCode;
	    }

		/**
		 * Checks if this graph equals an object.
		 *
		 * @param o	The object to compare against.
		 * @return True if both graphs are equal, i.e. same set of nodes and links, or else false.
	     * @throws NullPointerException If the object to compare against is false.
		 * @since jTLEX 1.0
		 */
	    @Override
	    public boolean equals(Object o) {
	        Objects.requireNonNull(o);

	        if (!(o instanceof ITimeMLGraph))
	            return false;

	        ITimeMLGraph obj = (TimeMLGraph) o;

	        if (this.getNodes().size() != obj.getNodes().size())
	            return false;

	        for (String nodeId : this.subNodes.keySet()) {
	            if (obj.getNodeById(nodeId) == null)
	                return false;
	        }

	        if (this.getLinks().size() != obj.getLinks().size())
	            return false;

	        for (int linkId : this.subLinks.keySet()) {
	            if (obj.getLinkById(linkId) == null)
	                return false;
	        }

	        return true;
	    }
		}
			  
		
		
	
}

