package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex;
import edu.fiu.jtlex.timeml.data.TimeMLLink;
import edu.fiu.jtlex.timeml.data.TimeMLTimex;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;

import java.util.*;

/**
 * Default implementation of {@link IDisconnectivityProcessor}
 *
 * @see IDisconnectivityProcessor
 * @since jTLEX 1.0
 */
public class DisconnectivityProcessor implements IDisconnectivityProcessor {

    /**
     * Creates a new instance
     *
     * @since jTLEX 1.0
     */
    public DisconnectivityProcessor() {}

    /**
     * Gets the document creation time
     *
     * @param partition The graph partition
     * @return The document creation time
     * @throws NullPointerException If the partition is null
     * @since jTLEX 1.0
     */
    @Override
    public TimeMLTimex getDct(ITimeMLGraph partition) {
        Objects.requireNonNull(partition);

        Optional<TimeMLTimex> timex = partition.getNodes().stream()
                .filter(node -> node instanceof TimeMLTimex)
                .map(t -> (TimeMLTimex) t)
                .filter(t -> t.getFunction() != null)
                .filter(this::timexIsDct)
                .findFirst();

        return timex.orElse(null);
    }

    /**
     * Checks if the Timex function equals the ITimex function
     *
     * @param timex to evaluate if the function in Document is equal to <ul><li>CREATION_TIME</li> or <li>PUBLICATION_TIME</li> </ul>
     *              timex can not be null
     * @return True if the timex function equals <ul><li>CREATION_TIME</li> or <li>PUBLICATION_TIME</li> </ul>
     * @since jTLEX 1.0
     */
    private boolean timexIsDct(TimeMLTimex timex) {
        return timex.getFunction().equals(ITimeMLTimex.FunctionInDocument.CREATION_TIME) ||
                timex.getFunction().equals(ITimeMLTimex.FunctionInDocument.PUBLICATION_TIME);
    }

    /**
     * Flags if a partition contains a document creation time.
     *
     * @param partition The graph partition.
     * @return True if the partition contains a document creation time, false if it doesn't.
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    @Override
    public boolean hasDct(ITimeMLGraph partition) {
        return getDct(partition) != null;
    }

    /**
     * Flags if the partition contains time expressions.
     *
     * @param partition the nodes are processed from the grab the time expressions.
     * @return True if the partition contains time expressions, false if it doesn't.
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    @Override
    public boolean hasTimeExpressions(ITimeMLGraph partition) {
        return partition.getNodes().stream().map(ITimeMLNode::getIdStr).anyMatch(str -> str.contains("t"));
    }

    /**
     * Gets a list of time expressions
     *
     * @param partition the nodes are processed from the grab the time expressions
     * @return A List of timexes
     * @throws NullPointerException If partition is null
     * @since jTLEX 1.0
     */
    @Override
    public List<TimeMLTimex> getTimeExpressions(ITimeMLGraph partition) {
        List<TimeMLTimex> timexes = partition.getNodes().stream()
                .filter(node -> node.getIdStr().contains("t"))
                .map(node -> (TimeMLTimex) node)
                .filter(timex -> timex.getType() != null)
                .filter(timex -> typeIsAllowed(timex.getType()))
                .filter(timex -> valueIsAllowed(timex.getValue().strip()))
                .toList();

        return timexes;
    }

    /**
     * Values rejected are <ul> <li>PXY</li><li>PAST_REF</li><li>PRESENT_RE</li>F<li>FUTURE_REF</li></ul>
     *
     * @since jTLEX 1.0
     */
    @SuppressWarnings("serial")
    private static final Set<String> NOT_ALLOWED_VALUES = new HashSet<>() {
        {
            add("PAST_REF");
            add("PRESENT_REF");
            add("FUTURE_REF");
            add("PXY");
        }
    };

    /**
     * Types not allowed are <ul><li>DURATION</li><li>SET type</li></ul>
     *
     * @since jTLEX 1.0
     */
    @SuppressWarnings("serial")
    private static final Set<ITimeMLTimex.TimexType> NOT_ALLOWED_TYPES = new HashSet<>() {
        {
            add(ITimeMLTimex.TimexType.DURATION);
            add(ITimeMLTimex.TimexType.SET);
        }
    };

    /**
     * Values allowed are all but the not allowed values of <ul> <li>PXY</li><li>PAST_REF</li><li>PRESENT_RE</li>F<li>FUTURE_REF</li></ul>
     *
     * @param value
     * @return true if it is an allowed value
     * @since jTLEX 1.0
     */
    private boolean valueIsAllowed(String value) {
        return !NOT_ALLOWED_VALUES.contains(value);
    }

    /**
     * Types allowed are all but the not allowed values of<ul><li>DURATION</li><li>SET type</li></ul>
     *
     * @param timexType type of timex
     * @return true if it is an allowed type
     * @since jTLEX 1.0
     */
    private boolean typeIsAllowed(ITimeMLTimex.TimexType timexType) {
        return !NOT_ALLOWED_TYPES.contains(timexType);
    }

    /**
     * Suggests a list of possible links that might connect the partition with date creation time
     * with the other partitions. Only one link per partitions pair is suggested.
     * <p>
     * TODO: Is it possible to keep track of which partitions the link is connecting?
     *
     * @param partitions A List of partitions
     * @param linksCount The number of links in the graph
     * @return A List with the suggested links to increase connectivity
     * @throws IllegalArgumentException If the graph does not have a date creation time tag.
     * @since jTLEX 1.0
     */
    @Override
    public List<ITimeMLLink> suggestLinks(List<ITimeMLGraph> partitions, int linksCount) {
        Objects.requireNonNull(partitions);

        ITimeMLGraph dctPartition = partitions.stream().filter(this::hasDct).findFirst().orElse(null);
        if (dctPartition == null)
            throw new IllegalArgumentException("The given graph does not have a date creation time");
        TimeMLTimex dct = getDct(dctPartition);

        List<ITimeMLGraph> candidates = partitions.stream()
                .filter(p -> !hasDct(p))
                .filter(this::hasTimeExpressions).toList();

        List<ITimeMLLink> suggestions = new ArrayList<>();

        for (ITimeMLGraph candidate : candidates) {
            List<TimeMLTimex> timexes = getTimeExpressions(candidate);
            ITimeMLLink suggestion = getSuggestion(dct, timexes, linksCount);
            if (suggestion != null)
                suggestions.add(suggestion);
        }

        return suggestions;
    }

    /**
     * Attempts to parse a suggested link verifying the tags
     *
     * @param dct             document creation time
     * @param timeExpressions the time expressions parsed
     * @param numberOfLinks   count of links
     * @return The link to be suggested
     * @throws NullPointerException if dct is null
     * @throws NullPointerException if timeExpressions is null
     * @since jTLEX 1.0
     */
    private ITimeMLLink getSuggestion(TimeMLTimex dct, List<TimeMLTimex> timeExpressions, int numberOfLinks) {
        Objects.requireNonNull(dct);
        Objects.requireNonNull(timeExpressions);

        List<TimeMLTimex> filteredTimexes = timeExpressions.stream().filter(t -> !t.getValue().contains("X")).toList();
        for (TimeMLTimex timex : filteredTimexes) {
            return decideByYear(dct, timex, numberOfLinks + 1);
        }

        return null;
    }

    /**
     * Creates a new link based on the document creation time year and timex year
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute year
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByYear(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {

        int dctYear = Integer.parseInt(dct.getValue().substring(0, 4));
        int timexYear = Integer.parseInt(timex.getValue().substring(0, 4));

        if (timexYear < dctYear) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.BEFORE, ITimeMLLink.Origintype.USER);
        } else if (timexYear > dctYear) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.AFTER, ITimeMLLink.Origintype.USER);
        } else {
            return decideByMonth(dct, timex, numberOfLinks);
        }
    }

    /**
     * Creates a new link based on the document creation time month and timex month
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute month
     * @param numberOfLinks count of link
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByMonth(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {
        String timexMonthStr = timex.getValue().length() >= 7
                ? timex.getValue().substring(5, 7)
                : "";

        if (isNotValid(timexMonthStr)) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", dct, null,
                    timex, ITimeMLLink.TLink.INCLUDES, ITimeMLLink.Origintype.USER);
        } else if (timexMonthStr.contains("Q")) {
            return decideByQuarter(dct, timex, numberOfLinks);
        } else if (timexMonthStr.contains("H")) {
            return decideByHalf(dct, timex, numberOfLinks);
        } else if (timexMonthStr.contains("W")) {
            return decideByWeek(dct, timex, numberOfLinks);
        }

        int dctMonth = Integer.parseInt(dct.getValue().substring(5, 7));
        int timexMonth = Integer.parseInt(timexMonthStr);

        if (timexMonth < dctMonth) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.BEFORE, ITimeMLLink.Origintype.USER);
        } else if (timexMonth > dctMonth) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.AFTER, ITimeMLLink.Origintype.USER);
        } else {
            return decideByDay(dct, timex, numberOfLinks);
        }
    }

    /**
     * Strings that are not valid include <ul><li>X</li><li>A</li><li>S</li><li>P</li><li>U</li><li>E</li><li>F</li><li>""</li></ul>
     *
     * @param string
     * @return TRUE if string contains <ul><li>X</li><li>A</li><li>S</li><li>P</li><li>U</li><li>E</li><li>F</li><li>""</li></ul>
     * @since jTLEX 1.0
     */
    private boolean isNotValid(String string) {
        return string.contains("X") || string.contains("A") || string.contains("S") ||
                string.contains("P") || string.contains("U") || string.contains("E") ||
                string.contains("F") || string.equals("");
    }

    /**
     * Creates a new link based on the document creation time week and timex week
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute week
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByWeek(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {
        Calendar calendar = Calendar.getInstance();

        int dctYear = Integer.parseInt(dct.getValue().substring(0, 4));
        int dctMonth = Integer.parseInt(dct.getValue().substring(5, 7));
        int dctDay = Integer.parseInt(dct.getValue().substring(8, 10));

        calendar.set(dctYear, dctMonth, dctDay);
        int weekOfYear = calendar.get(Calendar.WEEK_OF_YEAR);
        String dctWeek = "W" + weekOfYear;

        String timexWeek = timex.getValue().substring(5, 8);

        return decide(dct, dctWeek, timex, timexWeek, numberOfLinks);
    }

    /**
     * Creates a new link based on the document creation time quarter and timex quarter
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute quarter
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByQuarter(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {
        int dctMonth = Integer.parseInt(dct.getValue().substring(5, 7));
        String dctQuarter = convertMonthToQuarter(dctMonth);
        String timexQuarter = timex.getValue().substring(5, 7);

        return decide(dct, dctQuarter, timex, timexQuarter, numberOfLinks);
    }

    /**
     * Creates a new link based on the document creation time by half of the year and timex by half of the year
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute by half of the year
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByHalf(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {
        int dctMonth = Integer.parseInt(dct.getValue().substring(5, 7));
        String dctHalf = convertMonthToHalf(dctMonth);
        String timexHalf = timex.getValue().substring(5, 7);

        return decide(dct, dctHalf, timex, timexHalf, numberOfLinks);
    }

    /**
     * Creates a new link based on the document creation time day and timex day
     *
     * @param dct           document creation time
     * @param timex         timex used for the value attribute day
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decideByDay(TimeMLTimex dct, TimeMLTimex timex, int numberOfLinks) {
        String timexDay = timex.getValue().length() >= 10
                ? timex.getValue().substring(8, 10)
                : "";

        if (isNotValid(timexDay)) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", dct, null,
                    timex, ITimeMLLink.TLink.INCLUDES, ITimeMLLink.Origintype.USER);
        }

        String dctDay = dct.getValue().substring(8, 10);
        return decide(dct, dctDay, timex, timexDay, numberOfLinks);
    }

    /**
     * Creates a new link based on the comparison of dct value to timex value
     *
     * @param dct           document creation time
     * @param dctStr        document creating time string
     * @param timex         timex used for the value attribute of time
     * @param timexStr      timex value attribute as string
     * @param numberOfLinks count of links
     * @return new link
     * @throws NullPointerException if dct is null
     * @since jTLEX 1.0
     */
    private TimeMLLink decide(TimeMLTimex dct, String dctStr, TimeMLTimex timex, String timexStr, int numberOfLinks) {
        int comparisonResult = timexStr.compareTo(dctStr);

        if (comparisonResult < 0) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.BEFORE, ITimeMLLink.Origintype.USER);
        } else if (comparisonResult > 0) {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", timex, null,
                    dct, ITimeMLLink.TLink.AFTER, ITimeMLLink.Origintype.USER);
        } else {
            return new TimeMLLink(ITimeMLLink.TLink.class, numberOfLinks, "", dct, null,
                    timex, ITimeMLLink.TLink.INCLUDES, ITimeMLLink.Origintype.USER);
        }
    }

    /**
     * Gets the quarter the month is bounded to out of the 12 months.
     *
     * @param month Month ranging from <code>1-12</code>
     * @return The quarter the month is bounded to: Q1-Q2-Q3-Q4
     * @throws IllegalArgumentException If the month is out of range
     * @since jTLEX 1.0
     */
    @Override
    public String convertMonthToQuarter(int month) {
        switch (month) {
            case 1:
            case 2:
            case 3:
                return "Q1";
            case 4:
            case 5:
            case 6:
                return "Q2";
            case 7:
            case 8:
            case 9:
                return "Q3";
            case 10:
            case 11:
            case 12:
                return "Q4";
            default:
                throw new IllegalArgumentException("Month must be an integer between 1 and 12, inclusive");
        }
    }

    /**
     * Gets the half the month is bounded to out of the 12 months.
     *
     * @param month Month ranging from <code>1-12</code>
     * @return The half the month is bounded to: H1-H2.
     * @throws IllegalArgumentException If the month is out of range
     * @since jTLEX 1.0
     */
    @Override
    public String convertMonthToHalf(int month) {
        switch (month) {
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
                return "H1";
            case 7:
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:
                return "H2";
            default:
                throw new IllegalArgumentException("Month must be an integer between 1 and 12, inclusive");
        }
    }
}
