package edu.fiu.jtlex.algorithms;

import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraph;
import edu.fiu.jtlex.timeml.graph.ITimeMLGraphTCSP;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;
import edu.fiu.jtlex.timeml.graph.TimeMLGraph;
import edu.fiu.jtlex.timeml.graph.TimeMLGraphTCSP;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Default implementation for inconsistency detection
 *
 * @since jTLEX 1.0
 */
public class InconsistencyDetector implements IInconsistencyDetector {

    private final IGraphTCSPSolver _tcspSolver;

    /**
     * Creates a new indeterminacy detector with {@link GraphTCSPSolver}
     * as default implementation.
     * 
     * @since jTLEX 1.0
     */
    public InconsistencyDetector() {
        this(new GraphTCSPSolver());
    }

    /**
     * Creates a new inconsistency detector with the given {@link IGraphTCSPSolver} implementation.
     *
     * @param tcspSolver An {}solver implementation
     * @throws NullPointerException If the TCSP solver is null
     * @since jTLEX 1.0
     */
    public InconsistencyDetector(IGraphTCSPSolver tcspSolver) {
        Objects.requireNonNull(tcspSolver);

        _tcspSolver = tcspSolver;
    }

    /**
     * A method to determine a graph's consistency. This is done by creating a
     * TCSP for the graph ({@link TimeMLGraphTCSP}) presented by the graphs links and nodes
     * trying to solve it.
     *
     * Note that to produce a {@link TimeMLGraphTCSP} the given graph cannot contain
     * any S-Link. Consider partitioning the graph before checking for consistency.
     *
     * If there is a solution to the TCSP then the graph is consistent,
     * otherwise it is inconsistent.
     *
     * @see TimeMLGraphTCSP
     * @see ITimeMLGraphTCSP
     * @param timeMLGraph A timeML graph without S-Link.
     * @return true if the graph is consistent, else false
     * @throws NullPointerException If the graph is null
     * @throws IllegalStateException If the timeMLGraph contains any S-Link
     * @since jTLEX 1.0
     */
    @Override
    public boolean isConsistent(ITimeMLGraph timeMLGraph) {
        ITimeMLGraphTCSP problem = new TimeMLGraphTCSP(timeMLGraph);
        return _tcspSolver.solve(problem) != null;
    }

    /**
     * Evaluates whether a given link is a self loop.
     *
     * @param link Link to evaluate.
     * @return True if the link is a self loop, else False.
     * @throws NullPointerException If the link is null
     * @since jTLEX 1.0
     */
    private boolean isSelfLoop(ITimeMLLink link) {
        Objects.requireNonNull(link);

        return link.getEventInstance().equals(link.getRelatedToNode())
                && link.getRelType() != ITimeMLLink.TLink.SIMULTANEOUS
                && link.getRelType() != ITimeMLLink.TLink.IDENTITY;
    }

    /**
     * Given a set of links, identifies which are inconsistent self references,
     * and produces a subgraph for each instance.
     *
     * @param links A set of links.
     * @return A set of IGraphs, each containing inconsistent self looping nodes.
     * @throws NullPointerException If links is null.
     * @since jTLEX 1.0
     */
    public Set<ITimeMLGraph> generateSelfLoopingSubGraphs(Set<ITimeMLLink> links, Set<ITimeMLNode> nodes) {
        Objects.requireNonNull(links);
        Set<ITimeMLGraph> graphs = new HashSet<ITimeMLGraph>();
        for (ITimeMLLink link : links) {
        	if (isSelfLoop(link)) {
        		graphs.add(linkToGraph(link, nodes));
        	}
        }
        return graphs;
    }

    /**
     * Creates a graph containing only the immediate nodes from a given link.
     *
     * @param link Link from which to create the graph.
     * @return A graph containing only the immediate nodes to the link.
     * @throws NullPointerException If the link is null
     * @since jTLEX 1.0
     */
    private ITimeMLGraph linkToGraph(ITimeMLLink link, Set<ITimeMLNode> nodes) {
        Objects.requireNonNull(link);

        ITimeMLGraph graph = new TimeMLGraph();
        addLinkToGraph(graph, link, nodes);
        return graph;
    }

    /**
     * Adds a link and its immediate nodes to the given graph
     *
     * @param graph Graph to modify
     * @param link Link to add
     * @throws NullPointerException If either the graph or the link are null
     * @since jTLEX 1.0
     */
    private void addLinkToGraph(ITimeMLGraph graph, ITimeMLLink link, Set<ITimeMLNode> nodes) {
        Objects.requireNonNull(graph);
        Objects.requireNonNull(link);
  	   Map<String, ITimeMLNode> nodesMap = nodes.stream().collect(
               Collectors.toMap(node -> node.getIdStr(), node -> node));
  	   
        graph.addNode(nodesMap.get(link.getEventInstance()));
        graph.addNode(nodesMap.get(link.getRelatedToNode()));
        graph.addLink(link);
    }

    /**
     * Given a TimeML graph generates a set of subgraphs that contain all the links
     * and nodes involved in the inconsistencies. This is done in a four steps process:
     *
     * <ol>
     *      <li>
     *          Find the self looping links, generate a subgraph for each of them,
     *          and add it to the output set. For more details see: {@link #generateSelfLoopingSubGraphs(Set, Set)}
     *      </li>
     *      <li>
     *          Separate links into two partitions: consistent and inconsistent ones.
     *          This is done by adding each non self looping link to an empty graph
     *          and testing its consistency. If the added link makes the graph inconsistent,
     *          it is added to the inconsistent partition, else to the consistent one.
     *      </li>
     *      <li>
     *          For each link in the inconsistent partition, complete its cycle by adding
     *          it along with links from the consistent partition to an empty graph
     *          until the graph becomes inconsistent. Once the cycle is produced, add it to the
     *          output set. For more details see: {@link #completeInconsistentCycle(ITimeMLGraph, List, ITimeMLGraph)}
     *      </li>
     *      <li>
     *          Since some cycles can have shared edges, step 3 is repeated for each link
     *          in the inconsistent cycle that is not the initial inconsistent link.
     *      </li>
     * </ol>
     *
     * @param timeMLGraph TimeML graph to evaluate.
     * @return A set containing the inconsistent subgraphs.
     * @throws NullPointerException If the graph is null
     * @since jTLEX 1.0
     */
    @Override
    public Set<ITimeMLGraph> generateInconsistentSubgraphs(ITimeMLGraph timeMLGraph) {
        Objects.requireNonNull(timeMLGraph);

        if (isConsistent(timeMLGraph))
            return new HashSet<>();

        Set<ITimeMLGraph> inconsistentSubGraphs = generateSelfLoopingSubGraphs(timeMLGraph.getLinks(), timeMLGraph.getNodes());

        ITimeMLGraph subGraph = new TimeMLGraph();
        timeMLGraph.getNodes().forEach(subGraph::addNode);

        Set<ITimeMLLink> links = timeMLGraph.getLinks().stream().filter(l -> !isSelfLoop(l)).collect(Collectors.toSet());
        List<ITimeMLLink> inconsistentLinks = new ArrayList<>();
        List<ITimeMLLink> consistentLinks = new ArrayList<>();

        for (ITimeMLLink link : links) {
            subGraph.addLink(link);
            if (isConsistent(subGraph)) {
                consistentLinks.add(link);
            } else {
                subGraph.removeLink(link);
                inconsistentLinks.add(link);
            }
        }
        
        for (ITimeMLLink inconsistentLink : inconsistentLinks) {
            ITimeMLGraph inconsistentCycle = completeInconsistentCycle(
                    linkToGraph(inconsistentLink, timeMLGraph.getNodes()), consistentLinks, timeMLGraph
            );
            Objects.requireNonNull(inconsistentCycle);
            inconsistentSubGraphs.add(inconsistentCycle);

            Set<ITimeMLLink> cycleLinks = inconsistentCycle.getLinks();
            cycleLinks.remove(inconsistentLink);

            for (ITimeMLLink link : cycleLinks) {
                ITimeMLGraph sharedCycle = completeInconsistentCycle(
                        linkToGraph(inconsistentLink, timeMLGraph.getNodes()),
                        consistentLinks.stream().filter(l -> l.getLinkID() != link.getLinkID()).toList(), timeMLGraph
                );

                if (sharedCycle != null)
                    inconsistentSubGraphs.add(sharedCycle);
            }
        }

        return inconsistentSubGraphs;
    }

    /**
     * Finds a cycle in a graph by adding links one at the time, until the graph becomes inconsistent.
     *
     * This is step 3 of {@link #generateInconsistentSubgraphs(ITimeMLGraph)}
     *
     * @param inconsistentCycle The initial state of the graph
     * @param consistentLinks The list of links to try.
     * @return The inconsistent cycle if there is one, else null.
     */
    private ITimeMLGraph completeInconsistentCycle(ITimeMLGraph inconsistentCycle, List<ITimeMLLink> consistentLinks, ITimeMLGraph originalGraph) {
    	
        if (!isConsistent(inconsistentCycle))
            return inconsistentCycle;

        ITimeMLGraph subGraph = inconsistentCycle.clone();
        
        for (ITimeMLLink link: consistentLinks) {
            addLinkToGraph(subGraph, link, originalGraph.getNodes());
            
            if (!isConsistent(subGraph)) {
                addLinkToGraph(inconsistentCycle, link, originalGraph.getNodes());

                return completeInconsistentCycle(
                        inconsistentCycle,
                        consistentLinks.stream().filter(l -> l.getLinkID() != link.getLinkID()).toList(), originalGraph);
            }
        }
        
        return null;
    }
}
