package edu.fiu.jtlex.timeml.data;

import java.util.Set;

import edu.fiu.jtlex.timeml.graph.ITimeMLNode.IType;

/**
 * Represents the EVENT tag.<p>
 * 
 * We consider "events" a cover term for situations that happen or occur.
 * Events can be punctual (1-2) or last for a period of time (3-4). We also
 * consider as events those predicates describing states or circumstances in
 * which something obtains or holds true (5). Not all stative predicates will be
 * marked up, however. 
 * <ol>
 * <li>Ferdinand Magellan, a Portuguese explorer, first <b>reached</b> the islands 
 * 		in search of spices.
 * <li>A fresh flow of lava, gas and debris <b>erupted</b> there Saturday.
 * <li>11,024 people, including local Aeta aborigines, <b>were evacuated</b> to 18
 * 		disaster relief centers.
 * <li>We are <b>expecting</b> a major eruption, he said in a telephone interview
 * 		early today.
 * <li>Israel has been scrambling to buy more masks abroad, after a <b>shortage</b>
 * 		of several hundred thousand gas masks.
 * </ol>
 * @since jTLEX 1.0
 * @author arada002
 *
 */
public interface ITimeMLEvent {
	/**
	 * An enum that represents the seven different event classes an object may
	 * belong to. Event may belong to REPORTING, PERCEPTION, ASPECTUAL, I_ACTION,
	 * I_STATE, STATE, or OCCURENCE event class.
	 */
	public static enum EventClass implements IType {
		/**
		 * Describes the action of a person or an organization declaring something,
		 * narrating an event, informing about an event, etc.
		 */
		REPORTING, 
		
		/**
		 * Events involving the physical perception of another event. 
		 * Such events are typically expressed by verbs like: 
		 * see, watch, glimpse, behold, view, hear, listen, overhear.
		 */
		PERCEPTION, 
		
		/**
		 * Grammatical device of aspectual predication
		 */
		ASPECTUAL, 
		
		/**
		 * An I_ACTION is an Intensional Action. <p>An I ACTION introduces an event 
		 * argument (which must be in the text explicitly) describing an action 
		 * or situation from which we can infer something given its relation 
		 * with the I_ACTION. <p>Explicit performative predicates are also included here.
		 * Note that the I_ACTION class does not cover states.
		 */
		I_ACTION,
		
		/**
		 * This class includes states that refer to alternative or possible worlds, 
		 * (delimited by square brackets in the examples below), which can be 
		 * introduced by subordinated clauses, nominalizations, or untensed VPs.
		 */
		I_STATE, 
		
		/**
		 * States describe circumstances in which something obtains or holds true.
		 */
		STATE, 
		
		/**
		 * This class includes all the many other kinds of events describing something
		 * that happens or occurs in the world.
		 */
		OCCURRENCE
	}

	/**
	 * Returns the integer part of the eventId. 
	 * Each event has to be identified by a unique ID number and String. 
	 * @since jTLEX 1.0
	 * @return Integer part of the eventID. May not be 0 or negative.
	 */
	public int getId();

	/**
	 * Returns the full string representation of the eventID.
	 * Each event has to be identified by a unique ID number and String. 
	 * @since jTLEX 1.0
	 * @return Full string representation of the eventID. May not return <code>null</code>.
	 */
	public String getIdStr();

	/**
	 * Returns the stem of the event.
	 * @since jTLEX 1.0
	 * @return the stem of the event. May return <code>null</code> if stem is not included in
	 *         annotation
	 */
	public String getStem();
	
	/**
	 * Returns the event class to which the object belongs.
	 * @since jTLEX 1.0
	 * @return the event class to which the object belongs. 
	 * May not return <code>null</code>.
	 */
	public EventClass getEventClass();
	
	/**
	 * Returns the String Representation of the Event.
	 * @since jTLEX 1.0
	 * @return the info of an Event in String Format. 
	 */
	public String toString();
	
	/**
	 * Sets the related Instance for the event.
	 * May be null.
	 * 
	 * @param instance the {@link ITimeMLInstance} to connect to the event.
	 * @since jTLEX 1.0
	 */
	public void setInstance(ITimeMLInstance instance);
	
	/**
	 * Returns the IDs of the instances related to the Event.
	 * May be null
	 * @since jTLEX 1.0
	 * @return the int ID of the Instances related to the event. 
	 */
	public Set<Integer> getInstances();
	
	/**
	 * Returns the JSON (RFC 8259) representation of the event.
	 * @since jTLEX 1.0
	 * @return the JSON (RFC 8259) representation of the event.
	 */
	public String toJson();
	
	/**
	 * Returns a field-by-field copy of the event.
	 * @since jTLEX 1.0
	 * @return  a field-by-field copy of the event.
	 */
	public ITimeMLEvent clone();
	
	/**
	 * Generates a hash code for the event.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode();

	
	/**
	 * Checks if this event equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both events are equal, i.e. same ID and eventClass, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);

	

}
