package edu.fiu.jtlex.timeml.data;


import edu.fiu.jtlex.timeml.data.ITimeMLEvent.EventClass;
import edu.fiu.jtlex.timeml.graph.ITimeMLNode;

/**
 * Represents the MAKEINSTANCE tag.<p>
 * "We distinguish between event <b>tokens</b> and <b>event instances</b> or realisations �
 * MAKEINSTANCE creates the actual realisation of an event. The motivation is
 * examples like John taught on Monday and Tuesday, where one verb represents
 * two events. In order to be able to annotate such cases, it is necessary
 * to create two <b>instances</b> of taught, representing the two different events.
 * The tense, aspect, part of speech, polarity, and modality of a particular
 * event instance are also represented in this tag. As such, a sentence such 
 * as <i>John taught today but he might not tomorrow</i> can be adequately 
 * represented.
 * MAKEINSTANCEs are created <b>in addition</b> to the event annotation
 * (which marks up the event token).<p>
 * In most cases, only one MAKEINSTANCE is needed, and will be automatically
 * created by the annotation tool, so the annotator only has to insert
 * an <b>additional</b> MAKEINSTANCE for the second (and further, if necessary) instance
 * or realisation of an event."
 * <br><br>-TimeML Annotation Guidelines Version 1.2.1
 * @since jTLEX 1.0
 * @author arada002
 *
 */

public interface ITimeMLInstance extends ITimeMLNode{
	
	/**
	 * An enum that represents the different tenses of event instance phrases.
	 */
	public enum Tense {
		/**
		 * Past tense.
		 */
		PAST,
		
		/**
		 * Present tense.
		 */
		PRESENT,
		
		/**
		 * Future tense.
		 */
		FUTURE,
		
		/**
		 * Verbs preceeded by any other modal auxiliar: For example, <i>must, may, might, 
		 * 		can, could, should, and would.</i>
		 * In this case, the attributes in MAKEINSTANCE will be tense=�NONE� and pos=�VERB�.
		 */
		NONE,
		
		/**
		 * Infinitive: For Example, <i>Callahan plans to <b>release</b> the information today.</i>
		 */
		INFINITIVE,
		
		/**
		 * Present Participle: For example, <i>King Hussein arrived in Washington yesterday 
		 * 		<b>seeking</b> to gain support for a new initiative.</i>
		 */
		PRESPART,
		
		/**
		 * Past Participle: For example, <i>Steel plates <b>found</b> at the scene are 
		 * 		being examined by experts.</i>
		 */
		PASTPART
	}
	
	/**
	 * An enum that represents the different aspects of an event instance phrase.
	 */
	public enum Aspect {
		/**
		 * Expresses an ongoing action.
		 */
		PROGRESSIVE,
		
		/**
		 * Describes an action viewed as a simple whole
		 */
		PERFECTIVE,
		
		/**
		 * Combination of {@code PROGRESSIVE} and {@code PERFECTIVE}.
		 */
		PERFECTIVE_PROGRESSIVE,
		
		/**
		 * Null aspect.
		 */
		NONE
	}
	
	/**
	 * An enum that represents the different part of speech categories.
	 */
	public enum Pos {
		/**
		 * Adjective
		 */
		ADJECTIVE,
		
		/**
		 * Noun
		 */
		NOUN,
		
		/**
		 * Verb
		 */
		VERB,
		
		/**
		 * Preposition
		 */
		PREPOSITION,
		
		/**
		 * Other
		 */
		OTHER
	}
	
	/**
	 * An enum that indicates whether the event instance phrase is negated.
	 */
	public enum Polarity {
		/**
		 * Not negated.
		 */
		POS, 
		
		/**
		 * Negated.
		 */
		NEG
	}
	
	/**
	 * Returns the unique ID number eventInstanceID.
	 * 
	 * @return the unique ID number eventInstanceID.
	 */
	public int getId() ;
	
	/**
	 * Returns the ID of the event to which the instance is related.
	 * @return the ID of the event to which the instance is related. 
	 * @since jTLEX 1.0
	 */
	public String getEvent();
	
	/**
	 * Returns the tense of the verbal phrase.
	 * @return the tense of the verbal phrase. 
	 * 		   May not return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public Tense getTense();
	
	/**
	 * Returns the aspect of the verbal phrase.
	 * @return the aspect of the verbal phrase. 
	 * 		   May not return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public Aspect getAspect();
	
	/**
	 * Returns the syntactic category of the phrase (part of speech).
	 * @return the syntactic category of the phrase (part of speech). 
	 * 			May not return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public Pos getPos();
	
	/**
	 * Returns the polarity of the eventInstance.
	 * The polarity of an event instance is a required attribute 
	 * represented by the boolean attribute, polarity. <br>If it is set to 
	 * �NEG�, the event instance is negated. <br>If it is set to
	 * �POS� or not present in the annotation, the event instance is not 
	 * negated. 
	 * @return polarity: 
	 * 			NEG if the phrase is negated, 
	 * 			POS if the phrase is not negated.
	 * 		   	May not return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public Polarity getPolarity();
	
	/**
	 * The modality attribute is only present if there is a modal 
	 * word present that modifies the instance.
	 * @see <a href="https://en.wikipedia.org/wiki/Modality_(natural_language)">
	 * Modality (natural language)</a>
	 * @return the modal word that modifies the instance. 
	 * 		   May return <code>null</code>.
	 * @since jTLEX 1.0
	 */
	public String getModality();
	
	/**
	 * Returns the signal object, present if multiple event instances are created
	 * <code>null</code> if not defined, may not be less than 0
	 * 
	 * @return the signal object, present if multiple event instances are created.
	 * @since jTLEX 1.0
	 */
	public ITimeMLSignal getSignal();
	
	/**
	 * Returns the cardinality of the event instance. <br>
	 * Cardinality is only present if there is a signal that 
	 * provides its value.<br>
	 * There are also cases, where the annotator
	 * can either create as many MAKEINSTANCEs as motivated by the text or one
	 * MAKEINSTANCE which carries a cardinality value. Examples are the following:
	 * <ol>
	 * <li>John taught twice on Monday.
	 * <li>John taught 150 times last year.
	 * </ol>
	 * In the first example, the annotator can either create two MAKEINSTANCEs
	 * or one MAKEINSTANCE with cardinality two. In the second example, the only
	 * practical solution is to create one MAKEINSTANCE with cardinality 150.
	 * @see ITimeMLInstance
	 * @return Cardinality {@code 0} if not defined. 
	 * 			May not return less than {@code 0}.
	 * @since jTLEX 1.0
	 */
	public String getCardinality();
	
	/**
	 *Calls the event to which the instance is related, then returns the event class of the event instance.
	 *
	 *@return the event class of the event instance.
	 *@since jTLEX 1.0
	 */
	public EventClass getType();
	
	/**
	 * Grabs the eventInstanceID, adds eiid to the front for the String format and returns.
	 * 
	 * @return the eventInstanceID in string format.
	 * @since jTLEX 1.0
	 */
	public String getIdStr();
	
	
	/**
	 * Checks if this instance equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both instances are equal, i.e. same ID and event, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);
	

}

