package edu.fiu.jtlex.timeml.graph;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import edu.fiu.jtlex.timeml.data.ITimeMLEvent;
import edu.fiu.jtlex.timeml.data.ITimeMLLink;

/**
 * Abstraction of a TimeML Graph data structure
 * @since jTLEX 1.0
 */
public interface ITimeMLGraph extends Cloneable {
    /**
     * Refers to the different possible graph types.
     * @since jTLEX 1.0
     */
	public static enum GraphType {

	    /**
	     * Refers to the whole graph, which is composed by a set of partitions.
	     */
	    MainGraph("Main Graph"),

	    /**
	     * Refers to the partitions of the graph, which arise from the removal
	     * of S-Links
	     */
	    SubordinationGraph("Subordination Graph");
	    
	    private String name;

	    GraphType(String name) {
	        this.name = name;
	    }
	    
	    public String toString() {
	        return name;
	    }

	}

	/**
	 * Adds the provided link to the graph. The endpoints of the link must already
	 * be present in the Graph.
	 *
	 * @param link The ILink to be added.
	 * @return True if the link was added, otherwise it returns false if the link
	 *         already exists or if the nodes were not in the graph, and as a
	 *         result, the link is not added.
	 * @throws NullPointerException     if the link is null.
	 * @throws IllegalArgumentException if nodes are not in graph before link is
	 *                                  added.
	 * @since jTLEX 1.0
	 */
    boolean addLink(ITimeMLLink link);

	/**
	 * Adds a node to the graph.
	 *
	 * @param node The node to add
	 * @return True if the node was added, or False if the node already exists.
	 * @throws NullPointerException if the node is null.
	 * @since jTLEX 1.0
	 */
    boolean addNode(ITimeMLNode node);

	/**
	 * Gets a node by id.
	 *
	 * @param nodeId The id of the node to retrieve.
	 * @return The node with the given id, or null if no such node exist.
	 * @since jTLEX 1.0
	 */
    ITimeMLNode getNodeById(String nodeId);

	/**
	 * Gets a link by id.
	 *
	 * @param linkId The id of the link to retrieve.
	 * @return The link with the given id, or null if no such link exist.
	 * @since jTLEX 1.0
	 */
    ITimeMLLink getLinkById(int linkId);

    /**
	 * Gets a event by id.
	 * 
	 * @param eventId The id of the event to retrieve.
	 * @return The event with the given id, or null if no such event exists.
	 * @since jTLEX 1.0
	 */
    public ITimeMLEvent getEventById(String eventId);
    
	/**
	 * Removes a provided link from the graph.
	 *
	 * @param link The ILink to remove.
	 * @return true if removed, false if not removed or null input.
	 * @throws NullPointerException if the ILink is null.
	 * @since jTLEX 1.0
	 */
    boolean removeLink(ITimeMLLink link);

	/**
	 * Removes a provided node from the graph. Removes links it is involved with as
	 * well.
	 *
	 * @param node The node to remove.
	 * @return True if removed, false if not removed or null input.
	 * @throws NullPointerException if the node is null.
	 * @since jTLEX 1.0
	 */
    boolean removeNode(ITimeMLNode node);

	/**
	 * Removes a node by id. Removes links it is involved with as well.
	 *
	 * @param nodeId The id of the node to remove.
	 * @return True if removed, false if not removed.
	 * @since jTLEX 1.0
	 */
    boolean removeNodeById(String nodeId);

    
	/**
	 * Removes a link by id.
	 *
	 * @param linkId The id of the link to remove.
	 * @return True if removed, false if not removed or null input.
	 * @since jTLEX 1.0
	 */
    boolean removeLinkById(int linkId);
    
	/**
	 * Adds an event to the graph.
	 *
	 * @param event The event to add
	 * @return True if the event was added, or False if the event already exists.
     * @throws NullPointerException If the node is null
	 * @since jTLEX 1.0
	 */
    public boolean addEvent(ITimeMLEvent event);

	/**
	 * Removes a provided event from the graph. Removes instances it is involved with as
	 * well.
	 *
	 * @param event The event to remove.
	 * @return True if removed, false if not removed or null input.
	 * @throws NullPointerException if the node is null.
	 * @since jTLEX 1.0
	 */
    public boolean removeEvent(ITimeMLEvent event);

	/**
	 * Removes an event by id. Removes instances it is involved with as well.
	 *
	 * @param eventId The id of the event to remove.
	 * @return True if removed, false if not removed.
	 * @since jTLEX 1.0
	 */
    public boolean removeEventById(String eventId);
	/**
	 * Returns a snapshot of all the nodes currently in the graph. Modifying this
	 * set won't modify the actual graph's node set.
	 * 
	 * @return A set containing the nodes currently in the graph.
	 * @since jTLEX 1.0
	 */
    Set<ITimeMLNode> getNodes();

	/**
	 * Returns a snapshot of all the links currently in the graph. Modifying this
	 * set won't modify the actual graph's link set.
	 * 
	 * @return A set containing the links currently in the graph.
	 * @since jTLEX 1.0
	 */
    Set<ITimeMLLink> getLinks();
    
	/**
	 * Returns a snapshot of all the events currently in the graph. Modifying this
	 * set won't modify the actual graph's event set.
	 * 
	 * @return A set containing the events currently in the graph.
	 * @since jTLEX 1.0
	 */
    public Set<ITimeMLEvent> getEvents();
    

	/**
	 * Each link has a node at its beginning and end. As such, each node can have an
	 * incoming link or an outgoing link. This method is used to track each outgoing
	 * link from each node.
	 * 
	 * @return A map with all the nodes in the graph and their outgoing links.
	 * @since jTLEX 1.0
	 */
    public Map<String, Set<ITimeMLLink>> getOutLinks();
    
	/**
	 * Each link has a node at its beginning and end. As such, each node can have an
	 * incoming link or an outgoing link. This method is used to track each incoming
	 * link from each node.
	 *
	 * @return A map with all the nodes in the graph and their incoming links.
	 * @since jTLEX 1.0
	 */
    public Map<String, Set<ITimeMLLink>> getInLinks();

	/**
	 * Gets the Graph's type.
	 *
	 * @return A {@link  GraphType} representing the graph type.
	 * @since jTLEX 1.0
	 */
    GraphType getGraphType();
    
	/**
	 * Gets the Graph's Parent Graph. Used for obtaining the parent graph from
	 * Subordination Graphs.
	 *
	 * @return A TimeMLGraph representing the parent, or null if it is not a
	 *         subordination graph, meaning there is no parent.
	 * @since jTLEX 1.0
	 */
	ITimeMLGraph getParent();
    

	/**
	 * Creates a clone of the graph by first cloning the nodes and links and then returning 
	 * them inside a new Graph object.
	 *
	 * @return A field-by-field copy of the graph.
	 * @since jTLEX 1.0
	 */
    ITimeMLGraph clone();

	/**
	 * Returns the JSON (RFC 8259) representation of the graph.
	 *
	 * @return A flattened JSON string of the graph.
	 * @since jTLEX 1.0
	 */
    String toJson();
    
	/**
	 * Generates a hash code for the graph, based of its nodes and edges.
	 *
	 * @return The hash code.
	 * @since jTLEX 1.0
	 */
    @Override
    public int hashCode();
    
	/**
	 * Checks if this graph equals an object.
	 *
	 * @param o	The object to compare against.
	 * @return True if both graphs are equal, i.e. same set of nodes and links, or else false.
     * @throws NullPointerException If the object to compare against is false.
	 * @since jTLEX 1.0
	 */
    @Override
    public boolean equals(Object o);
    

	/**
	 * Creates a sub graph.
	 * 
	 * @param nodes	The set of {@link ITimeMLNode}.
	 * @param links	The set of {@link ITimeMLLink}.
	 * @param events The set of {@link ITimeMLEvent}.
	 * @return an ITimeMLGraph that is the subgraph.
	 * @throws IllegalArgumentException if the nodes or links are not inside the graph
	 * the subgraph is being created from 
	 * @since jTLEX 1.0
	 */
	ITimeMLGraph subGraph(Set<ITimeMLNode> nodes, Set<ITimeMLLink> links, Set<ITimeMLEvent> events) throws IllegalArgumentException;
	
	
	
	/**
	 * Gets the link count per link type.
	 *
	 * @param graph	the TimeML Graph to count the links from.
	 * @return A map containing the number of A-Link, S-Link, and T-Link.
	 * @since jTLEX 1.0
	 */
    public static Map<Class<?>, Integer> getLinkTypesCount(ITimeMLGraph graph) {
        
        Map<Class<?>, Integer> LinkTypes = new HashMap<>();

        Set<ITimeMLLink> links = graph.getLinks();
        int ALinks = 0;
        int SLinks = 0;
        int TLinks = 0;

        for (ITimeMLLink L : links) {
            if (L.getRelType() instanceof ITimeMLLink.ALink) {
                ALinks++;
            } else if (L.getRelType() instanceof ITimeMLLink.SLink) {
                SLinks++;
            } else if (L.getRelType() instanceof ITimeMLLink.TLink) {
                TLinks++;
            }
        }
        
        LinkTypes.put(ITimeMLLink.ALink.class, ALinks);
        //Change .LinkTag to class input
        LinkTypes.put(ITimeMLLink.SLink.class, SLinks);
        LinkTypes.put(ITimeMLLink.TLink.class, TLinks);
        return LinkTypes;
    }

}
