package edu.fiu.jtlex.timeml.io;
import edu.fiu.jtlex.timeml.data.ITimeMLEvent;
import edu.fiu.jtlex.timeml.data.ITimeMLInstance;
import edu.fiu.jtlex.timeml.data.ITimeMLSignal;
import edu.fiu.jtlex.timeml.data.ITimeMLLink;
import edu.fiu.jtlex.timeml.data.ITimeMLText;
import edu.fiu.jtlex.timeml.data.ITimeMLTimex;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;


/**
* The interface for the master parser to allow for different parts of
* a timeML document (Event, Timex, and Link) to be parsed for the required
* variables within their strings.
* 
* @since jTLEX 1.0
* @author ceber003
*
*/
public interface ITimeMLParser {

	
	   /**
		* The parse method which will take the input stream and break it up into 
		* the different variables held within.
	    * @throws UnsupportedEncodingException the text must be encoded using UFT-8.
	    * @throws NullPointerException if tml argument is <code>null</code>
	    * @param timeML the TimeML annotated text file to be parsed.
	    * @return a new Text object which contains all of the separated data.
	    * @since jTLEX 1.0
		*/		
	   public ITimeMLText parse(InputStream timeML) throws UnsupportedEncodingException;
		
		
	   /**
	    * A parser method which takes the string input representing an Event and parses
	    * it to get the different variables from the string. 
	    * May be empty or whitespace however it will raise an IllegalArgumentException when 
	    * creating the Event object.
	    *
	    * @param eventString a string input representing an Event.
	    * @return a new Event object.
	    * @throws NullPointerException if null
	    * @since jTLEX 1.0
	    */	
	   public ITimeMLEvent parseEvent(String eventString);
		

	   /**
	    * A parser method which takes the string input representing a Timex and parses
	    * it to get the different variables from the string.
	    * May be empty or whitespace however it will raise an NullPointerException when 
	    * creating the Timex object.
	    *
	    * @param timexString a string input representing a time expression.
	    * @return a new Timex object.
	    * @throws NullPointerException if null
	    * @since jTLEX 1.0
	    */	
	   public ITimeMLTimex parseTimex(String timexString);
		
		
	   /**
	    * A parser method which takes the string input representing a Link and parses
	    * it to get the different variables from the string.
	    * May be empty or whitespace however it will raise an NullPointerException when 
	    * creating the Link object.
	    *  
	    * @param tag an XML tag relating to the link.
	    * @param linkString a string input representing a link.
	    * @return a new Link object.
	    * @throws NullPointerException if null
	    * @since jTLEX 1.0
	    */	
	   public ITimeMLLink parseLink(String tag, String linkString);
	   
	   /**
	    * A parser method which takes the string input representing an Instance and parses
	    * it to get the Instance object. 
	    * May be empty or whitespace however it will raise an IllegalArgumentException when 
	    * creating the Instance object.
	    *
	    * @param instanceString a string input representing an instance.
	    * @return a new Instance object.
	    * @throws NullPointerException if null
	    * @since jTLEX 1.0
	    */	
	   public ITimeMLInstance parseInstance(String instanceString);
	   
	   /**
	    * A parser method which takes the string input representing a Signal and parses
	    * it to get the Signal object. 
	    * May be empty or whitespace however it will raise an IllegalArgumentException when 
	    * creating the Signal object.
	    *
	    * @param signalString a string representation of a signal.
	    * @return a new Signal object.
	    * @throws NullPointerException if null
	    * @since jTLEX 1.0
	    */	
	   public ITimeMLSignal parseSignal(String signalString);
		
	   
}
