package edu.fiu.jtlex.timeml.timeline;

import org.jacop.core.IntVar;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * Default implementation of {@code IIndeterminacy}.
 *
 * @since jTLEX 1.0
 */
public class TimeMLIndeterminacy implements ITimeMLIndeterminacy {

    private final List<IntVar> _shortestTimeLine;
    private final List<Boolean> _indeterminatePairs;
    private final Set<Integer> _indeterminateIndexes;

	/**
	 * Creates a new Indeterminacy with the specified information.
	 * 
	 * @param shortestTimeLine	a list of int variables containing the shortest timeline.
	 * @param indeterminatePairs	a list of booleans with the indeterminate pairs.
	 * @param indeterminateIndexes	a set of integers with the indexes that are indeterminate
	 * @throws NullPointerException     if shortestTimeLine, indeterminatePairs, or indeterminateIndexes are null.
	 */
    public TimeMLIndeterminacy(
            List<IntVar> shortestTimeLine, List<Boolean> indeterminatePairs, Set<Integer> indeterminateIndexes) {
        Objects.requireNonNull(shortestTimeLine);
        Objects.requireNonNull(indeterminatePairs);
        Objects.requireNonNull(indeterminateIndexes);

        _shortestTimeLine = shortestTimeLine;
        _indeterminatePairs = indeterminatePairs;
        _indeterminateIndexes = indeterminateIndexes;
    }

    /**
     * Calculates the indeterminacy score.
     *
     * @return A number between 0-1 representing the indeterminacy score
     * @since jTLEX 1.0
     */
    @Override
    public double getIndeterminacyScore() {
        double indeterminatePairCount = _indeterminatePairs.stream().filter(Boolean::booleanValue).count();
        return indeterminatePairCount / _indeterminatePairs.size();
    }

    /**
     * Return the adjacent indeterminant time pairings.
     *
     * @return A List of the adjacent indeterminant time pairings.
     * @since jTLEX 1.0
     */
    @Override
    public List<String> indeterminantTimePairs() {
        List<String> indeterminatePairsList = new ArrayList<>();

        for (int i = 0; i < _indeterminatePairs.size(); i++) {
            if (_indeterminatePairs.get(i)) {
                String pair = _shortestTimeLine.get(i).id() + "&" + _shortestTimeLine.get(i + 1).id();
                indeterminatePairsList.add(pair);
            }
        }

        return indeterminatePairsList;
    }

    /**
     * Returns the indeterminant time points
     *
     * @return A List of the indeterminate time points as Strings.
     * @since jTLEX 1.0
     */
    @Override
    public List<String> indeterminantTimePoints() {
        List<String> indeterminateTimePointsList = new ArrayList<>();

        for (int index : _indeterminateIndexes) {
            indeterminateTimePointsList.add(_shortestTimeLine.get(index).id());
        }

        return indeterminateTimePointsList;
    }

    /**
     * Returns the total number of indeterminant time sections.
     *
     * @return The total number of indeterminate time sections.
     * @since jTLEX 1.0
     */
    @Override
    public int getTotalIndeterminantSections() {
        return _indeterminateIndexes.size();
    }

    /**
     * Returns the timeline's length.
     *
     * @return The timeline's length.
     * @since jTLEX 1.0
     */
    @Override
    public int getTimelineLength() {
        return _shortestTimeLine.size();
    }

    /**
     * Returns the number of adjacent pairs.
     *
     * @return The number of adjacent pairs.
     * @since jTLEX 1.0
     */
    @Override
    public int getTotalAdjacentPairs() {
        return (int)_indeterminatePairs.stream().filter(Boolean::booleanValue).count();
    }
}
